# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""HTML viewer tab widget"""

import os.path
from utils.fileutils import getFileContent
from utils.colorfont import getZoomedMonoFont
from .qt import QFrame, QHBoxLayout, QTextBrowser, Qt, pyqtSignal
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase


class HTMLViewer(QTextBrowser):

    """HTML viewer (web browser)"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, parent=None):
        QTextBrowser.__init__(self, parent)

    def keyPressEvent(self, event):
        """Handles the key press events"""
        if event.key() == Qt.Key_Escape:
            self.sigEscapePressed.emit()
            event.accept()
        else:
            QTextBrowser.keyPressEvent(self, event)

    def onTextZoomChanged(self):
        """Triggered when a text zoom is changed.
           It is mostly used in diff viewers
        """
        self.setFont(getZoomedMonoFont())


class HTMLTabWidget(MainWindowTabWidgetBase, QFrame):

    """The widget which displays a RO HTML page"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, parent=None):

        MainWindowTabWidgetBase.__init__(self)
        QFrame.__init__(self, parent)

        self.setFrameShape(QFrame.StyledPanel)
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.__editor = HTMLViewer(self)
        self.__editor.sigEscapePressed.connect(self.__onEsc)
        layout.addWidget(self.__editor)

        self.__fileName = ""
        self.__shortName = ""
        self.__encoding = "n/a"

    def __onEsc(self):
        """Triggered when Esc is pressed"""
        self.sigEscapePressed.emit()

    def setHTML(self, content):
        """Sets the content from the given string"""
        self.__editor.setHtml(content)

    def getHTML(self):
        """Provides the currently shown HTML"""
        return self.__editor.toHtml()

    def loadFormFile(self, path):
        """Loads the content from the given file"""
        self.setHTML(getFileContent(path))
        self.__fileName = path
        self.__shortName = os.path.basename(path)

    def onTextZoomChanged(self):
        """Triggered when a text zoom is changed"""
        self.__editor.onTextZoomChanged()

    def getViewer(self):
        """Provides the QWebView"""
        return self.__editor

    def setFocus(self):
        """Overridden setFocus"""
        self.__editor.setFocus()

    def isModified(self):
        """Tells if the file is modifed"""
        return False

    def getRWMode(self):
        """Tells the read/write mode"""
        return "RO"

    def getType(self):
        """Tells the widget type"""
        return MainWindowTabWidgetBase.HTMLViewer

    def getLanguage(self):
        """Tells the content language"""
        return "HTML"

    def getFileName(self):
        """Tells what file name of the widget"""
        return self.__fileName

    def setFileName(self, path):
        """Sets the file name"""
        self.__fileName = path
        self.__shortName = os.path.basename(path)

    def getEncoding(self):
        """Tells the content encoding"""
        return self.__encoding

    def setEncoding(self, newEncoding):
        """Sets the encoding - used for Diff files"""
        self.__encoding = newEncoding

    def getShortName(self):
        """Tells the display name"""
        return self.__shortName

    def setShortName(self, name):
        """Sets the display name"""
        self.__shortName = name



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""The file outline viewer implementation"""

import os.path
import logging
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.settings import Settings
from utils.fileutils import isPythonMime, isPythonFile
from cdmpyparser import getBriefModuleInfoFromMemory
from .qt import (Qt, QSize, QTimer, QCursor, QPalette, QMenu, QWidget, QAction,
                 QVBoxLayout, QToolBar, QFrame, QLabel)
from .outlinebrowser import OutlineBrowser
from .viewitems import (FunctionItemType, ClassItemType, AttributeItemType,
                        GlobalItemType)
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase
from .parsererrors import ParserErrorsDialog


class OutlineAttributes:

    """Holds all the attributes associated with an outline browser"""

    def __init__(self):
        self.browser = None
        self.contextItem = None
        self.info = None
        self.shortFileName = ""
        self.changed = False


class FileOutlineViewer(QWidget):

    """ The file outline viewer widget """

    def __init__(self, editorsManager, parent=None):
        QWidget.__init__(self, parent)

        self.__editorsManager = editorsManager
        self.__mainWindow = parent
        self.__editorsManager.currentChanged.connect(self.__onTabChanged)
        self.__editorsManager.sigTabClosed.connect(self.__onTabClosed)
        self.__editorsManager.sigBufferSavedAs.connect(self.__onSavedBufferAs)
        self.__editorsManager.sigFileTypeChanged.connect(
            self.__onFileTypeChanged)

        self.__outlineBrowsers = {}  # UUID -> OutlineAttributes
        self.__currentUUID = None
        self.__updateTimer = QTimer(self)
        self.__updateTimer.setSingleShot(True)
        self.__updateTimer.timeout.connect(self.__updateView)

        self.findButton = None
        self.outlineViewer = None
        self.toolbar = None
        self.__createLayout()

        self.__modifiedFormat = Settings()['modifiedFormat']

        # create the context menu
        self.__menu = QMenu(self)
        self.__findMenuItem = self.__menu.addAction(
            getIcon('findusage.png'), 'Find where used', self.__findWhereUsed)

    def setTooltips(self, switchOn):
        """Sets the tooltips mode"""
        for key in self.__outlineBrowsers:
            self.__outlineBrowsers[key].browser.setTooltips(switchOn)

    def __connectOutlineBrowser(self, browser):
        """Connects a new buffer signals"""
        browser.setContextMenuPolicy(Qt.CustomContextMenu)
        browser.customContextMenuRequested.connect(
            self.__handleShowContextMenu)
        browser.sigFirstSelectedItem.connect(self.__selectionChanged)

    def __createLayout(self):
        """Helper to create the viewer layout"""
        # Toolbar part - buttons
        self.findButton = QAction(
            getIcon('findusage.png'), 'Find where highlighted item is used',
            self)
        self.findButton.setVisible(False)
        self.findButton.triggered.connect(self.__findWhereUsed)
        self.showParsingErrorsButton = QAction(
            getIcon('showparsingerrors.png'), 'Show lexer/parser errors', self)
        self.showParsingErrorsButton.triggered.connect(self.__showParserError)
        self.showParsingErrorsButton.setEnabled(False)

        self.toolbar = QToolBar(self)
        self.toolbar.setMovable(False)
        self.toolbar.setAllowedAreas(Qt.TopToolBarArea)
        self.toolbar.setIconSize(QSize(16, 16))
        self.toolbar.setFixedHeight(28)
        self.toolbar.setContentsMargins(0, 0, 0, 0)
        self.toolbar.addAction(self.findButton)
        self.toolbar.addAction(self.showParsingErrorsButton)

        # Prepare members for reuse
        self.__noneLabel = QLabel("\nNot a python file")
        self.__noneLabel.setFrameShape(QFrame.StyledPanel)
        self.__noneLabel.setAlignment(Qt.AlignHCenter)
        headerFont = self.__noneLabel.font()
        headerFont.setPointSize(headerFont.pointSize() + 2)
        self.__noneLabel.setFont(headerFont)
        self.__noneLabel.setAutoFillBackground(True)
        noneLabelPalette = self.__noneLabel.palette()
        noneLabelPalette.setColor(QPalette.Background,
                                  GlobalData().skin['nolexerPaper'])
        self.__noneLabel.setPalette(noneLabelPalette)

        self.__layout = QVBoxLayout()
        self.__layout.setContentsMargins(0, 0, 0, 0)
        self.__layout.setSpacing(0)
        self.__layout.addWidget(self.toolbar)
        self.__layout.addWidget(self.__noneLabel)

        self.setLayout(self.__layout)

    def __selectionChanged(self, index):
        """Handles the changed selection"""
        if index.isValid():
            self.__outlineBrowsers[self.__currentUUID].contentItem = \
                self.__outlineBrowsers[
                    self.__currentUUID].browser.model().item(index)
        else:
            self.__outlineBrowsers[self.__currentUUID].contentItem = None
        self.__updateButtons()

    def __handleShowContextMenu(self, coord):
        """ Show the context menu """
        browser = self.__outlineBrowsers[self.__currentUUID].browser
        index = browser.indexAt(coord)
        if not index.isValid():
            return

        # This will update the contextItem
        self.__selectionChanged(index)

        contextItem = self.__outlineBrowsers[self.__currentUUID].contentItem
        if contextItem:
            self.__findMenuItem.setEnabled(self.findButton.isEnabled())
            self.__menu.popup(QCursor.pos())

    def __goToDefinition(self):
        """Jump to definition context menu handler"""
        contextItem = self.__outlineBrowsers[self.__currentUUID].contentItem
        if contextItem is not None:
            self.__outlineBrowsers[
                self.__currentUUID].browser.openItem(contextItem)

    def __findWhereUsed(self):
        """Find where used context menu handler"""
        contextItem = self.__outlineBrowsers[self.__currentUUID].contentItem
        if contextItem is not None:
            GlobalData().mainWindow.findWhereUsed(
                contextItem.getPath(),
                contextItem.sourceObj)

    def __updateButtons(self):
        """Updates the toolbar buttons depending on what is selected"""
        self.findButton.setEnabled(False)

        contextItem = self.__outlineBrowsers[self.__currentUUID].contentItem
        if contextItem is None:
            if contextItem.itemType in [FunctionItemType, ClassItemType,
                                        AttributeItemType, GlobalItemType]:
                self.findButton.setEnabled(True)

    def __onTabChanged(self, index):
        """Triggered when another tab becomes active"""
        # If the timer is still active that means the tab was switched before
        # the handler had a chance to work. Therefore update the previous tab
        # first if so.
        if self.__updateTimer.isActive():
            self.__updateTimer.stop()
            self.__updateView()

        # Now, switch the outline browser to the new tab
        if index == -1:
            widget = self.__editorsManager.currentWidget()
        else:
            widget = self.__editorsManager.getWidgetByIndex(index)
        if widget is None:
            self.__currentUUID = None
            self.__noneLabel.show()
            self.showParsingErrorsButton.setEnabled(False)
            return
        if widget.getType() not in [MainWindowTabWidgetBase.PlainTextEditor,
                                    MainWindowTabWidgetBase.VCSAnnotateViewer]:
            if self.__currentUUID is not None:
                self.__outlineBrowsers[self.__currentUUID].browser.hide()
                self.__currentUUID = None
            self.__noneLabel.show()
            self.showParsingErrorsButton.setEnabled(False)
            return

        # This is text editor, detect the file type
        if not isPythonMime(widget.getMime()):
            if self.__currentUUID is not None:
                self.__outlineBrowsers[self.__currentUUID].browser.hide()
                self.__currentUUID = None
            self.__noneLabel.show()
            self.showParsingErrorsButton.setEnabled(False)
            return

        # This is a python file, check if we already have the parsed info in
        # the cache
        uuid = widget.getUUID()
        if uuid in self.__outlineBrowsers:
            # We have it, hide the current and show the existed
            if self.__currentUUID is not None:
                self.__outlineBrowsers[self.__currentUUID].browser.hide()
                self.__currentUUID = None
            else:
                self.__noneLabel.hide()
            self.__currentUUID = uuid
            self.__outlineBrowsers[self.__currentUUID].browser.show()

            info = self.__outlineBrowsers[self.__currentUUID].info
            self.showParsingErrorsButton.setEnabled(info.isOK != True)
            return

        # It is first time we are here, create a new
        editor = widget.getEditor()
        editor.textChanged.connect(self.__onBufferChanged)
        editor.cursorPositionChanged.connect(self.__cursorPositionChanged)
        info = getBriefModuleInfoFromMemory(editor.text)

        self.showParsingErrorsButton.setEnabled(info.isOK != True)

        shortFileName = widget.getShortName()
        browser = OutlineBrowser(uuid, shortFileName, info, self)
        browser.setHeaderHighlight(info.isOK != True)
        self.__connectOutlineBrowser(browser)
        self.__layout.addWidget(browser)
        if self.__currentUUID is not None:
            self.__outlineBrowsers[self.__currentUUID].browser.hide()
            self.__currentUUID = None
        else:
            self.__noneLabel.hide()

        self.__currentUUID = uuid
        attributes = OutlineAttributes()
        attributes.browser = browser
        attributes.contextItem = None
        attributes.info = info
        attributes.shortFileName = shortFileName
        attributes.changed = False
        self.__outlineBrowsers[self.__currentUUID] = attributes
        self.__outlineBrowsers[self.__currentUUID].browser.show()

    def getCurrentUsedInfo(self):
        """Provides the info used to show the current outline window"""
        if self.__currentUUID in self.__outlineBrowsers:
            return self.__outlineBrowsers[self.__currentUUID].info
        return None

    def __cursorPositionChanged(self):
        """Triggered when a cursor position is changed"""
        if self.__updateTimer.isActive():
            # If a file is very large and the cursor is moved
            # straight after changes this will delay the update till
            # the real pause.
            self.__updateTimer.stop()
            self.__updateTimer.start(1500)

    def __onBufferChanged(self):
        """Triggered when a change in the buffer is identified"""
        if self.__currentUUID is None:
            return
        widget = self.__editorsManager.getWidgetByUUID(self.__currentUUID)
        if widget is None:
            return
        if self.__mainWindow.debugMode:
            return

        self.__updateTimer.stop()
        if self.__currentUUID in self.__outlineBrowsers:
            if not self.__outlineBrowsers[self.__currentUUID].changed:
                self.__outlineBrowsers[self.__currentUUID].changed = True
                browser = self.__outlineBrowsers[self.__currentUUID].browser
                fName = self.__outlineBrowsers[
                    self.__currentUUID].shortFileName
                title = self.__modifiedFormat % fName
                browser.model().sourceModel().updateRootData(0, title)
        self.__updateTimer.start(1500)

    def __updateView(self):
        """Updates the view when a file is changed"""
        self.__updateTimer.stop()
        info = self.getCurrentBufferInfo()
        if info is None:
            return

        self.showParsingErrorsButton.setEnabled(info.isOK != True)
        browser = self.__outlineBrowsers[self.__currentUUID].browser
        fName = self.__outlineBrowsers[self.__currentUUID].shortFileName
        browser.setHeaderHighlight(info.isOK != True)

        if not info.isOK:
            title = self.__modifiedFormat % fName
            browser.model().sourceModel().updateRootData(0, title)
            return

        browser.model().sourceModel().updateRootData(0, fName)
        self.__outlineBrowsers[self.__currentUUID].changed = False

        browser.updateFileItem(browser.model().sourceModel().rootItem, info)
        self.__outlineBrowsers[self.__currentUUID].info = info

    def getCurrentBufferInfo(self):
        """Provides the current buffer parsed info"""
        if self.__currentUUID is None:
            return None
        widget = self.__editorsManager.getWidgetByUUID(self.__currentUUID)
        if widget is None:
            return None

        editor = widget.getEditor()
        info = getBriefModuleInfoFromMemory(editor.text)
        return info

    def __onTabClosed(self, uuid):
        """Triggered when a tab is closed"""
        if uuid in self.__outlineBrowsers:
            browser = self.__outlineBrowsers[uuid].browser
            self.__layout.removeWidget(browser)
            browser.hide()
            browser.deleteLater()
            del self.__outlineBrowsers[uuid]
            if uuid == self.__currentUUID:
                self.__currentUUID = None

    def __onSavedBufferAs(self, fileName, uuid):
        """Triggered when a file is saved with a new name"""
        if uuid in self.__outlineBrowsers:
            baseName = os.path.basename(fileName)
            if not isPythonFile(fileName):
                # It's not a python file anymore
                if uuid == self.__currentUUID:
                    self.__outlineBrowsers[uuid].browser.hide()
                    self.__noneLabel.show()
                    self.__currentUUID = None

                del self.__outlineBrowsers[uuid]
                self.showParsingErrorsButton.setEnabled(False)
                self.findButton.setEnabled(False)
                return

            # Still python file with a different name
            browser = self.__outlineBrowsers[uuid].browser
            self.__outlineBrowsers[uuid].shortFileName = baseName
            if self.__outlineBrowsers[uuid].changed:
                title = self.__modifiedFormat % baseName
            else:
                title = baseName
            browser.model().sourceModel().updateRootData(0, title)

    def __onFileTypeChanged(self, fileName, uuid, newFileType):
        """Triggered when the current buffer file type is changed, e.g. .cgi"""
        del fileName    # unused argument
        if isPythonMime(newFileType):
            # The file became a python one
            if uuid not in self.__outlineBrowsers:
                self.__onTabChanged(-1)
        else:
            if uuid in self.__outlineBrowsers:
                # It's not a python file any more
                if uuid == self.__currentUUID:
                    self.__outlineBrowsers[uuid].browser.hide()
                    self.__noneLabel.show()
                    self.__currentUUID = None

                del self.__outlineBrowsers[uuid]
                self.showParsingErrorsButton.setEnabled(False)
                self.findButton.setEnabled(False)

    def __showParserError(self):
        """Shows the parser errors window"""
        if self.__currentUUID is None:
            return

        try:
            widget = self.__editorsManager.getWidgetByUUID(self.__currentUUID)
            if widget is None:
                return

            editor = widget.getEditor()
            info = getBriefModuleInfoFromMemory(editor.text)
            fName = self.__outlineBrowsers[self.__currentUUID].shortFileName
            dialog = ParserErrorsDialog(fName, info)
            dialog.exec_()
        except Exception as ex:
            logging.error(str(ex))

    def highlightContextItem(self, context, line):
        """Highlights the context item"""
        if not self.__currentUUID in self.__outlineBrowsers:
            return False
        browser = self.__outlineBrowsers[self.__currentUUID].browser
        info = self.__outlineBrowsers[self.__currentUUID].info
        return browser.highlightContextItem(context, line, info)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""qutepart text editor component wrapper"""


import re
import logging
import os.path
import socket
import urllib.request
from qutepart import Qutepart
from ui.qt import (QPalette, pyqtSignal, QTextCursor, QApplication,
                   QCursor, Qt, QDesktopServices, QUrl)
from utils.globals import GlobalData
from utils.settings import Settings
from utils.colorfont import getZoomedMonoFont
from utils.encoding import decodeURLContent


WORD_AT_END_REGEXP = re.compile("\w+$")
WORD_AT_START_REGEXP = re.compile("^\w+")


class QutepartWrapper(Qutepart):

    """Convenience qutepart wrapper"""

    sigHighlighted = pyqtSignal(str, int, int)

    # Shared between buffers
    matchesRegexp = None
    highlightOn = False

    def __init__(self, parent):
        Qutepart.__init__(self, needMarkArea=False, needLineNumbers=False,
                          needCompleter=False)
        self.setParent(parent)

        self.encoding = None
        self.explicitUserEncoding = None
        self.mime = None

        # Search/replace support
        self.__matchesCache = None
        self.textChanged.connect(self.__resetMatchCache)

    def _dropUserExtraSelections(self):
        """Suppressing highlight removal when the text is changed"""
        pass

    def setPaper(self, paperColor):
        """Sets the new paper color"""
        palette = self.palette()
        palette.setColor(QPalette.Active, QPalette.Base, paperColor)
        palette.setColor(QPalette.Inactive, QPalette.Base, paperColor)
        self.setPalette(palette)

    def setColor(self, textColor):
        """Sets the new text color"""
        palette = self.palette()
        palette.setColor(QPalette.Active, QPalette.Text, textColor)
        palette.setColor(QPalette.Inactive, QPalette.Text, textColor)
        self.setPalette(palette)

    def onTextZoomChanged(self):
        """Triggered when a text zoom is changed"""
        self.setFont(getZoomedMonoFont())
        for margin in self.getMargins():
            if hasattr(margin, 'onTextZoomChanged'):
                margin.onTextZoomChanged()
        self._setSolidEdgeGeometry()

    def clearUndoRedoHistory(self):
        """Clears the undo/redo history"""
        self.document().clearUndoRedoStacks()

    def getEolIndicator(self):
        """Provides the eol indicator for the current eol mode"""
        if self.eol == '\r\n':
            return "CRLF"
        if self.eol == '\r':
            return 'CR'
        return 'LF'

    def firstVisibleLine(self):
        """Provides the first visible line. 0-based"""
        return self.firstVisibleBlock().blockNumber()

    def setFirstVisible(self, lineno):
        """Scrolls the editor to make sure the first visible line is lineno"""
        currentVisible = self.firstVisibleLine()
        if currentVisible == lineno:
            return

        # Initial setting
        self.verticalScrollBar().setValue(lineno)
        currentVisible = self.firstVisibleLine()

        while currentVisible != lineno:
            vbValue = self.verticalScrollBar().value()
            distance = lineno - currentVisible
            if distance > 0:
                distance = min(2, distance)
            else:
                distance = max(-2, distance)
            self.verticalScrollBar().setValue(vbValue + distance)
            vbValueAfter = self.verticalScrollBar().value()
            if vbValueAfter == vbValue:
                break
            currentVisible = self.firstVisibleLine()
        self.setHScrollOffset(0)

    def lastVisibleLine(self):
        """Provides the last visible line. 0-based"""
        editorHeight = self.height()
        hBar = self.horizontalScrollBar()
        if hBar:
            if hBar.isVisible():
                editorHeight -= hBar.height()
        block = self.firstVisibleBlock()

        lastVisible = block.blockNumber()
        blocksHeight = 0.0
        while block.isValid():
            if not block.isValid():
                break
            blocksHeight += self.blockBoundingRect(block).height()
            if blocksHeight > editorHeight:
                break
            lastVisible = block.blockNumber()
            block = block.next()
        return lastVisible

    def isLineOnScreen(self, line):
        """True if the line is on screen. line is 0-based."""
        if line < self.firstVisibleLine():
            return False
        return line <= self.lastVisibleLine()

    def ensureLineOnScreen(self, line):
        """Makes sure the line is visible on screen. line is 0-based."""
        # Prerequisite: the cursor has to be on the desired position
        if not self.isLineOnScreen(line):
            self.ensureCursorVisible()

    def setHScrollOffset(self, value):
        """Sets the new horizontal scroll bar value"""
        hBar = self.horizontalScrollBar()
        if hBar:
            hBar.setValue(value)

    def moveToLineEnd(self):
        """Moves the cursor to the end of the line"""
        line, _ = self.cursorPosition
        self.cursorPosition = line, len(self.lines[line])

    @staticmethod
    def firstNonSpaceIndex(text):
        """Provides a pos (0-based of a first non-space char in the text"""
        lStripped = text.lstrip()
        if lStripped:
            return len(text) - len(lStripped)
        return None

    def __getNewHomePos(self, toFirstNonSpace):
        """Provides the new cursor position for a HOME click"""
        line, pos = self.cursorPosition
        newPos = 0
        if toFirstNonSpace:
            lStripped = self.lines[line].lstrip()
            if lStripped:
                calcPos = len(self.lines[line]) - len(lStripped)
                newPos = 0 if pos <= calcPos else calcPos
        return line, newPos

    def moveToLineBegin(self, toFirstNonSpace):
        """Jumps to the first non-space or to position 0"""
        newLine, newPos = self.__getNewHomePos(toFirstNonSpace)
        self.cursorPosition = newLine, newPos

    def selectTillLineBegin(self, toFirstNonSpace):
        """Selects consistently with HOME behavior"""
        newLine, newPos = self.__getNewHomePos(toFirstNonSpace)
        cursor = self.textCursor()
        cursor.setPosition(self.mapToAbsPosition(newLine, newPos),
                           QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def onHome(self):
        """Triggered when HOME is received"""
        self.moveToLineBegin(Settings()['jumpToFirstNonSpace'])

    def onShiftHome(self):
        """Triggered when Shift+HOME is received"""
        self.selectTillLineBegin(Settings()['jumpToFirstNonSpace'])

    def onShiftEnd(self):
        """Selects till the end of line"""
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def onShiftDel(self):
        """Triggered when Shift+Del is received"""
        if self.selectedText:
            QApplication.clipboard().setText(self.selectedText)
            self.selectedText = ''
        else:
            line, _ = self.cursorPosition
            if self.lines[line]:
                QApplication.clipboard().setText(self.lines[line] + '\n')
                del self.lines[line]

    def onCtrlC(self):
        """Handles copying"""
        if self.selectedText:
            QApplication.clipboard().setText(self.selectedText)
        else:
            line, _ = self.cursorPosition
            if self.lines[line]:
                QApplication.clipboard().setText(self.lines[line] + '\n')

    def openAsFile(self):
        """Opens a selection or a current tag as a file"""
        path = self.selectedText.strip()
        if path == "" or '\n' in path or '\r' in path:
            return

        # Now the processing
        if os.path.isabs(path):
            GlobalData().mainWindow.detectTypeAndOpenFile(path)
            return

        # This is not an absolute path but could be a relative path for the
        # current buffer file. Let's try it.
        fileName = self._parent.getFileName()
        if fileName != "":
            # There is a file name
            fName = os.path.dirname(fileName) + os.path.sep + path
            fName = os.path.abspath(os.path.realpath(fName))
            if os.path.exists(fName):
                GlobalData().mainWindow.detectTypeAndOpenFile(fName)
                return
        if GlobalData().project.isLoaded():
            # Try it as a relative path to the project
            prjFile = GlobalData().project.fileName
            fName = os.path.dirname(prjFile) + os.path.sep + path
            fName = os.path.abspath(os.path.realpath(fName))
            if os.path.exists(fName):
                GlobalData().mainWindow.detectTypeAndOpenFile(fName)
                return
        # The last hope - open as is
        if os.path.exists(path):
            path = os.path.abspath(os.path.realpath(path))
            GlobalData().mainWindow.detectTypeAndOpenFile(path)
            return

        logging.error("Cannot find '" + path + "' to open")

    def downloadAndShow(self):
        """Triggered when the user wants to download and see the file"""
        url = self.selectedText.strip()
        if url.lower().startswith("www."):
            url = "http://" + url

        oldTimeout = socket.getdefaulttimeout()
        newTimeout = 5      # Otherwise the pause is too long
        socket.setdefaulttimeout(newTimeout)
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))

        try:
            response = urllib.request.urlopen(url)
            content = decodeURLContent(response.read())

            # The content has been read sucessfully
            mainWindow = GlobalData().mainWindow
            mainWindow.editorsManager().newTabClicked(content,
                                                      os.path.basename(url))
        except Exception as exc:
            logging.error("Error downloading '" + url + "'\n" + str(exc))

        QApplication.restoreOverrideCursor()
        socket.setdefaulttimeout(oldTimeout)

    def openInBrowser(self):
        """Triggered when a selected URL should be opened in a browser"""
        url = self.selectedText.strip()
        if url.lower().startswith("www."):
            url = "http://" + url
        QDesktopServices.openUrl(QUrl(url))

    def printUserData(self):
        """Debug purpose member to print the highlight data"""
        line, pos = self.cursorPosition
        if self._highlighter is None:
            print(str(line+1) + ":" + str(pos+1) + " no highlight")
            return
        block = self.document().findBlockByNumber(line)
        data = block.userData()
        if data is None:
            print(str(line+1) + ":" + str(pos+1) + " None")
            return
        print(str(line+1) + ":" + str(pos+1) + " " + repr(data.data))

    def removeTrailingWhitespaces(self):
        """Removes trailing whitespaces"""
        # Note: two loops are for the case when there is nothing to expand.
        # The 'with' statement leads to a detection of a changed position
        # which triggers the other actions.
        index = 0
        for index, line in enumerate(self.lines):
            stripped = line.rstrip()
            if stripped != line:
                break
        else:
            self.__showStatusBarMessage('No trailing spaces found')
            return

        with self:
            lastIndex = len(self.lines) - 1
            while index <= lastIndex:
                stripped = self.lines[index].rstrip()
                if stripped != self.lines[index]:
                    self.lines[index] = stripped
                index += 1

    def expandTabs(self, tabsize):
        """Expands tabs if needed"""
        # Note: two loops are for the case when there is nothing to expand.
        # The 'with' statement leads to a detection of a changed position
        # which triggers the other actions.
        index = 0
        for index, line in enumerate(self.lines):
            expanded = line.expandtabs(tabsize)
            if expanded != line:
                break
        else:
            self.__showStatusBarMessage('No tabs found')
            return

        with self:
            lastIndex = len(self.lines) - 1
            while index <= lastIndex:
                expanded = self.lines[index].expandtabs(tabsize)
                if expanded != self.lines[index]:
                    self.lines[index] = expanded
                index += 1

    def getEncoding(self):
        """Provides the encoding"""
        if self.explicitUserEncoding:
            return self.explicitUserEncoding
        return self.encoding

    def isCommentLine(self, line):
        """True if it is a comment line. line is 0-based"""
        if line >= len(self.lines):
            return False
        txt = self.lines[line]
        nonSpaceIndex = self.firstNonSpaceIndex(txt)
        if nonSpaceIndex is None:
            return False
        if txt[nonSpaceIndex] != '#':
            return False
        return not self.isStringLiteral(line, nonSpaceIndex)

    def isLineEmpty(self, line):
        """Returns True if the line is empty. Line is 0 based"""
        return self.lines[line].strip() == ""

    def isStringLiteral(self, line, pos):
        """True if it is a string literal"""
        if self._highlighter is None:
            return False
        if pos < 0:
            # May happened if the line is empty
            return False

        block = self.document().findBlockByNumber(line)
        data = block.userData()
        if data is None:
            return False
        try:
            return self._highlighter._syntax._getTextType(data.data,
                                                          pos) == 's'
        except IndexError:
            return False

    # Search supporting members

    def resetHighlight(self):
        """Resets the highlight if so"""
        self.resetMatchCache()
        self.setExtraSelections([])
        QutepartWrapper.highlightOn = False

    def __resetMatchCache(self):
        """Resets the matches cache when the text is modified"""
        # The highlight does not need to be switched off
        self.__matchesCache = None

    def resetMatchCache(self):
        """Resets the matches cache.

        Happens when a search criteria changed in the otehr editors.
        """
        if not self.isVisible():
            if self._userExtraSelections:
                self.setExtraSelections([])
        self.__matchesCache = None

    def __searchInText(self, regExp, startPoint, forward):
        """Search in text and return the nearest match"""
        self.findAllMatches(regExp)
        if self.__matchesCache:
            if forward:
                for match in self.__matchesCache:
                    if match.start() >= startPoint:
                        break
                else:  # wrap, search from start
                    match = self.__matchesCache[0]
            else:  # reverse search
                for match in self.__matchesCache[::-1]:
                    if match.start() <= startPoint:
                        break
                else:  # wrap, search from end
                    match = self.__matchesCache[-1]
            return match
        return None

    def isCursorOnMatch(self):
        """True if the cursor is on the first pos of any match"""
        if self.__matchesCache:
            pos = self.absCursorPosition
            for match in self.__matchesCache:
                if match.start() == pos:
                    return True
        return False

    def getCurrentMatchesCount(self):
        """Provides the number of the current matches"""
        if self.__matchesCache:
            return len(self.__matchesCache)
        return 0

    def getMatchesInfo(self):
        """Returns match number or None and total number of matches"""
        matchNumber = None
        totalMatches = None
        if self.__matchesCache:
            pos = self.absCursorPosition
            totalMatches = 0
            for match in self.__matchesCache:
                totalMatches += 1
                if match.start() == pos:
                    matchNumber = totalMatches
        return matchNumber, totalMatches

    def getCurrentOrSelection(self):
        """Provides what should be used for search.

        Returns a tuple:
        - word
        - True if it was a selection
        - start abs pos
        - end abs pos
        """
        cursor = self.textCursor()
        if cursor.hasSelection():
            word = cursor.selectedText()
            if '\r' not in word and '\n' not in word:
                return word, True, cursor.anchor(), cursor.position()
        cursor.select(QTextCursor.WordUnderCursor)
        return cursor.selectedText(), False, cursor.anchor(), cursor.position()

    def getCurrentWord(self):
        """Provides the current word if so"""
        cursor = self.textCursor()
        cursor.select(QTextCursor.WordUnderCursor)
        return cursor.selectedText()

    def getWordBeforeCursor(self):
        """Provides a word before the cursor"""
        cursor = self.textCursor()
        textBeforeCursor = cursor.block().text()[:cursor.positionInBlock()]
        match = WORD_AT_END_REGEXP.search(textBeforeCursor)
        return match.group(0) if match else ''

    def getWordAfterCursor(self):
        """Provides a word after cursor"""
        cursor = self.textCursor()
        textAfterCursor = cursor.block().text()[cursor.positionInBlock():]
        match = WORD_AT_START_REGEXP.search(textAfterCursor)
        return match.group(0) if match else ''

    def onFirstChar(self):
        """Jump to the first character in the buffer"""
        self.cursorPosition = 0, 0
        self.ensureLineOnScreen(0)
        self.setHScrollOffset(0)

    def onLastChar(self):
        """Jump to the last char"""
        line = len(self.lines)
        if line != 0:
            line -= 1
        pos = len(self.lines[line])
        self.cursorPosition = line, pos
        self.ensureLineOnScreen(line)
        self.setHScrollOffset(0)

    def clearSelection(self):
        """Clears the current selection if so"""
        cursor = self.textCursor()
        cursor.clearSelection()

    def findAllMatches(self, regExp):
        """Find all matches of regExp"""
        if QutepartWrapper.matchesRegexp != regExp or \
                self.__matchesCache is None:
            QutepartWrapper.matchesRegexp = regExp
            self.__matchesCache = [match
                                   for match in regExp.finditer(self.text)]
        QutepartWrapper.highlightOn = True
        return self.__matchesCache

    def updateFoundItemsHighlighting(self, regExp):
        """Updates the highlight. Returns False if there were too many."""
        matches = self.findAllMatches(regExp)
        count = len(matches)
        if count > Settings()['maxHighlightedMatches']:
            self.setExtraSelections([])
            return False

        self.setExtraSelections([(match.start(), len(match.group(0)))
                                 for match in matches])
        return True

    def highlightRegexp(self, regExp, searchPos, forward, needMessage=True):
        """Highlights the matches, moves cursor, displays message"""
        highlighted = self.updateFoundItemsHighlighting(regExp)
        match = self.__searchInText(regExp, searchPos, forward)
        if match is not None:
            matchIndex = self.__matchesCache.index(match) + 1
            totalMatches = len(self.__matchesCache)
            self.absCursorPosition = match.start()
            self.ensureCursorVisible()

        if needMessage:
            if highlighted:
                if self.__matchesCache:
                    msg = 'match %d of %d' % (matchIndex, totalMatches)
                else:
                    msg = 'no matches'
            else:
                msg = 'match %d of %d (too many to highlight, ' \
                      'exceeds the limit of %d)' % \
                      (matchIndex, totalMatches,
                       Settings()['maxHighlightedMatches'])
            self.__showStatusBarMessage(msg)
        return len(self.__matchesCache)

    def clearSearchIndicators(self):
        """Hides the search indicator"""
        self.resetHighlight()
        GlobalData().mainWindow.clearStatusBarMessage()

    def onHighlight(self):
        """Triggered when Ctrl+' is clicked"""
        if self.isCursorOnHighlight():
            return self.onNextHighlight()

        word, wasSelection, _, absEnd = self.getCurrentOrSelection()
        if not word or '\r' in word or '\n' in word:
            return 0

        # Reset match cashe in all the buffers
        # Otherwise they keep an old highlight when the user switches to them
        mainWindow = GlobalData().mainWindow
        mainWindow.editorsManager().resetTextSearchMatchCache()

        wordFlag = 0
        if wasSelection:
            regExp = re.compile('%s' % re.escape(word), re.IGNORECASE)
        else:
            regExp = re.compile('\\b%s\\b' % re.escape(word), re.IGNORECASE)
            wordFlag = 1

        count = self.highlightRegexp(regExp, absEnd, False)
        self.sigHighlighted.emit(word, wordFlag, count)
        return count

    def mouseDoubleClickEvent(self, event):
        """Highlight the current word keeping the selection"""
        Qutepart.mouseDoubleClickEvent(self, event)
        if self.selectedText:
            selectedPos = self.absSelectedPosition
            self.onHighlight()
            self.absSelectedPosition = selectedPos

    def onNextHighlight(self):
        """Triggered when Ctrl+. is clicked"""
        if QutepartWrapper.matchesRegexp is None:
            return self.onHighlight()

        if QutepartWrapper.highlightOn:
            # The current highlight is on
            return self.highlightRegexp(QutepartWrapper.matchesRegexp,
                                        self.absCursorPosition + 1, True)

        # Not highlighted. If within the currently matched word, then the
        # cursor should stay on it.
        wordBefore = self.getWordBeforeCursor()
        return self.highlightRegexp(QutepartWrapper.matchesRegexp,
                                    self.absCursorPosition - len(wordBefore),
                                    True)

    def onPrevHighlight(self):
        """Triggered when Ctrl+, is clicked"""
        if QutepartWrapper.matchesRegexp is None:
            return self.onHighlight()

        if QutepartWrapper.highlightOn:
            # The current highlight is on
            return self.highlightRegexp(QutepartWrapper.matchesRegexp,
                                        self.absCursorPosition - 1, False)

        # Not highlighted. If within the currently matched word, then the
        # cursor should stay on it.
        return self.highlightRegexp(QutepartWrapper.matchesRegexp,
                                    self.absCursorPosition, False)

    def replaceAllMatches(self, replaceText):
        """Replaces all the current matches with the other text"""
        if not self.__matchesCache:
            return

        replaceCount = 0
        noReplaceCount = 0
        for match in self.__matchesCache[::-1]:
            textToReplace = self.text[match.start():
                                      match.start() + len(match.group(0))]
            if textToReplace == replaceText:
                noReplaceCount += 1
            else:
                replaceCount += 1

        if replaceCount > 0:
            cursorPos = None
            delta = 0
            regExp = QutepartWrapper.matchesRegexp
            with self:
                # reverse order, because replacement may move indexes
                for match in self.__matchesCache[::-1]:
                    textToReplace = self.text[match.start():
                                              match.start() +
                                              len(match.group(0))]
                    if textToReplace != replaceText:
                        self.replaceText(match.start(), len(match.group(0)),
                                         replaceText)

                    if cursorPos is None:
                        cursorPos = self.absCursorPosition
                    else:
                        delta += len(replaceText) - len(textToReplace)

            self.resetHighlight()
            self.updateFoundItemsHighlighting(regExp)
            self.absCursorPosition = cursorPos + delta

        if replaceCount == 1:
            msg = '1 match replaced'
        else:
            msg = '%d matches replaced' % replaceCount

        if noReplaceCount > 0:
            msg += '; %d skipped ' \
                   '(the highlight matches replacement)' % noReplaceCount

        self.__showStatusBarMessage(msg)

    def replaceMatch(self, replaceText):
        """Replaces the match on which the cursor is"""
        if self.__matchesCache:
            pos = self.absCursorPosition
            for match in self.__matchesCache:
                if match.start() == pos:
                    regExp = QutepartWrapper.matchesRegexp
                    textToReplace = self.text[match.start():
                                              match.start() +
                                              len(match.group(0))]
                    if textToReplace == replaceText:
                        msg = "no replace: the highlight matches replacement"
                    else:
                        self.replaceText(match.start(), len(match.group(0)),
                                         replaceText)
                        self.__matchesCache = None
                        self.updateFoundItemsHighlighting(regExp)
                        msg = "1 match replaced"

                    self.__showStatusBarMessage(msg)
                    break

    def isCursorOnHighlight(self):
        """True if the current cursor position is on the highlighted text"""
        if QutepartWrapper.highlightOn:
            pos = self.absCursorPosition
            self.findAllMatches(QutepartWrapper.matchesRegexp)
            for match in self.__matchesCache:
                if pos >= match.start() and pos < match.end():
                    return True
        return False

    def onTabChanged(self):
        """Called by find/replace widget"""
        if QutepartWrapper.highlightOn:
            # Highlight the current regexp without changing the cursor position
            self.updateFoundItemsHighlighting(QutepartWrapper.matchesRegexp)
        else:
            self.resetHighlight()

    @staticmethod
    def __showStatusBarMessage(msg):
        """Shows a main window status bar message"""
        mainWindow = GlobalData().mainWindow
        mainWindow.showStatusBarMessage(msg, 8000)

    def getEndPosition(self):
        """Provides the end position, 0 based"""
        line = len(self.lines) - 1
        return (line, len(self.lines[line]))

    def append(self, text):
        """Appends the given text to the end"""
        if not text:
            return

        # Tail separator could be stripped otherwise; also there are so many
        # separators:
        # https://docs.python.org/3.5/library/stdtypes.html#str.splitlines
        partsNoEnd = text.splitlines()
        partsWithEnd = text.splitlines(True)

        lastIndex = len(partsNoEnd) - 1
        with self:
            for index, value in enumerate(partsNoEnd):
                if value:
                    self.lines[-1] += value
                if index == lastIndex:
                    if value != partsWithEnd[index]:
                        self.lines.append('')
                else:
                    self.lines.append('')

    def gotoLine(self, line, pos=None, firstVisible=None):
        """Jumps to the given position and scrolls if needed.

        line and pos and firstVisible are 1-based
        """
        # Normalize editor line and pos
        editorLine = line - 1
        if editorLine < 0:
            editorLine = 0
        if pos is None or pos <= 0:
            editorPos = 0
        else:
            editorPos = pos - 1

        if self.isLineOnScreen(editorLine):
            if firstVisible is None:
                self.cursorPosition = editorLine, editorPos
                return

        self.ensureLineOnScreen(editorLine)

        # Otherwise we would deal with scrolling any way, so normalize
        # the first visible line
        if firstVisible is None:
            editorFirstVisible = editorLine - 1
        else:
            editorFirstVisible = firstVisible - 1
        if editorFirstVisible < 0:
            editorFirstVisible = 0

        self.cursorPosition = editorLine, editorPos
        self.setFirstVisible(editorFirstVisible)

    def openAsFileAvailable(self):
        """True if there is something to try to open as a file"""
        selectedText = self.selectedText.strip()
        if selectedText:
            return '\n' not in selectedText and '\r' not in selectedText
        return False

    def downloadAndShowAvailable(self):
        """True if download and show available"""
        selectedText = self.selectedText.strip()
        if '\n' not in selectedText and '\r' not in selectedText:
            return selectedText.lower().startswith('http://') or \
                   selectedText.lower().startswith('www.')
        return False

    def terminate(self):
        """Cleans up memory"""
        Qutepart.terminate(self)
        self.document().disconnect()

        self._indenter._qpart = None
        self._indenter = None
        if self._highlighter is not None:
            self._highlighter._syntax = None
            self._highlighter._document = None
            self._highlighter._textEdit = None
        self._solidEdgeLine.deleteLater()
        self._solidEdgeLine = None
        self._rectangularSelection._qpart = None
        self._rectangularSelection = None
        self._lines._qpart = None
        self._lines._doc = None
        self._lines = None




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 6 and adopted for codimension.
# Original copyright:
# Copyright (c) 2016 - 2017 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Module implementing classes and functions to dump variable contents"""

#
# This code was inspired by pydevd.
#

MAX_ITEMS_TO_HANDLE = 300
TOO_LARGE_MESSAGE = ("Too large to show contents. Max items to show: " +
                     str(MAX_ITEMS_TO_HANDLE))
TOO_LARGE_ATTRIBUTE = "Too large to be handled"


#
# Classes implementing resolvers for various compund types
#


class BaseResolver(object):

    """Base class of the resolver class tree"""

    def resolve(self, var, attribute):
        """Provides an attribute from a variable"""
        raise NotImplementedError

    def getDictionary(self, var):
        """Provides the attributes of a variable as a dictionary"""
        raise NotImplementedError


#
# Default Resolver
#
class DefaultResolver(BaseResolver):

    """Class used to resolve the default way"""

    def resolve(self, var, attribute):
        """Provides an attribute from a variable"""
        return getattr(var, attribute, None)

    def getDictionary(self, var):
        """Provides the attributes of a variable as a dictionary"""
        names = dir(var)
        if not names and hasattr(var, "__members__"):
            names = var.__members__

        retVal = {}
        for name in names:
            try:
                attribute = getattr(var, name)
                retVal[name] = attribute
            except Exception:
                pass    # if we can't get it, simply ignore it
        return retVal


#
# Resolver for Dictionaries
#
class DictResolver(BaseResolver):

    """Class used to resolve from a dictionary"""

    def resolve(self, var, attribute):
        """Provides an attribute from a variable"""
        if attribute in ('___len___', TOO_LARGE_ATTRIBUTE):
            return None

        if "(ID:" not in attribute:
            try:
                return var[attribute]
            except Exception:
                return getattr(var, attribute, None)

        expectedID = int(attribute.split("(ID:")[-1][:-1])
        for key, value in var.items():
            if id(key) == expectedID:
                return value

        return None

    @staticmethod
    def keyToStr(key):
        """Provides a string representation for a key"""
        if isinstance(key, str):
            return repr(key)
        return key

    def getDictionary(self, var):
        """Provides the attributes of a variable as a dictionary"""
        retVal = {}
        count = 0
        for key, value in var.items():
            count += 1
            key = "{0} (ID:{1})".format(self.keyToStr(key), id(key))
            retVal[key] = value
            if count > MAX_ITEMS_TO_HANDLE:
                retVal[TOO_LARGE_ATTRIBUTE] = TOO_LARGE_MESSAGE
                break

        retVal["___len___"] = len(var)

        # in case it has additional fields
        additionals = defaultResolver.getDictionary(var)
        retVal.update(additionals)
        return retVal


#
# Resolver for Lists and Tuples
#
class ListResolver(BaseResolver):

    """Class used to resolve from a tuple or list"""

    def resolve(self, var, attribute):
        """Provides an attribute from a variable"""
        if attribute in ('___len___', TOO_LARGE_ATTRIBUTE):
            return None

        try:
            return var[int(attribute)]
        except Exception:
            return getattr(var, attribute, None)

    def getDictionary(self, var):
        """Provides the attributes of a variable as a dictionary"""
        retVal = {}
        count = 0
        for value in var:
            retVal[str(count)] = value
            count += 1
            if count > MAX_ITEMS_TO_HANDLE:
                retVal[TOO_LARGE_ATTRIBUTE] = TOO_LARGE_MESSAGE
                break

        retVal["___len___"] = len(var)

        # in case it has additional fields
        additionals = defaultResolver.getDictionary(var)
        retVal.update(additionals)
        return retVal


#
# Resolver for Sets and Frozensets
#
class SetResolver(BaseResolver):

    """Class used to resolve from a set or frozenset"""

    def resolve(self, var, attribute):
        """Provides an attribute from a variable"""
        if attribute in ('___len___', TOO_LARGE_ATTRIBUTE):
            return None

        if attribute.startswith("ID: "):
            attribute = attribute.split(None, 1)[1]
        try:
            attribute = int(attribute)
        except Exception:
            return getattr(var, attribute, None)

        for varAttr in var:
            if id(varAttr) == attribute:
                return varAttr
        return None

    def getDictionary(self, var):
        """Provides the attributes of a variable as a dictionary"""
        retVal = {}
        count = 0
        for value in var:
            count += 1
            retVal["ID: " + str(id(value))] = value
            if count > MAX_ITEMS_TO_HANDLE:
                retVal[TOO_LARGE_ATTRIBUTE] = TOO_LARGE_MESSAGE
                break

        retVal["___len___"] = len(var)

        # in case it has additional fields
        additionals = defaultResolver.getDictionary(var)
        retVal.update(additionals)
        return retVal


#
# Resolver for Numpy Arrays
#
class NdArrayResolver(BaseResolver):

    """Class used to resolve from numpy ndarray including some meta data"""

    @staticmethod
    def __isNumeric(arr):
        """Checks if an array is of a numeric type"""
        try:
            return arr.dtype.kind in 'biufc'
        except AttributeError:
            return False

    def resolve(self, var, attribute):
        """Provides an attribute from a variable"""
        if attribute == '__internals__':
            return defaultResolver.getDictionary(var)

        if attribute == 'min':
            if self.__isNumeric(var):
                return var.min()
            return None

        if attribute == 'max':
            if self.__isNumeric(var):
                return var.max()
            return None

        if attribute == 'mean':
            if self.__isNumeric(var):
                return var.mean()
            return None

        if attribute == 'shape':
            return var.shape

        if attribute == 'dtype':
            return var.dtype

        if attribute == 'size':
            return var.size

        if attribute.startswith('['):
            container = NdArrayItemsContainer()
            count = 0
            for element in var:
                setattr(container, str(count), element)
                count += 1
                if count > MAX_ITEMS_TO_HANDLE:
                    setattr(container, TOO_LARGE_ATTRIBUTE, TOO_LARGE_MESSAGE)
                    break
            return container
        return None

    def getDictionary(self, var):
        """Provides the attributes of a variable as a dictionary"""
        retVal = {}
        retVal['__internals__'] = defaultResolver.getDictionary(var)
        if var.size > 1024 * 1024:
            retVal['min'] = 'ndarray too big, calculating min would ' \
                            'slow down debugging'
            retVal['max'] = 'ndarray too big, calculating max would ' \
                            'slow down debugging'
        else:
            if self.__isNumeric(var):
                retVal['min'] = var.min()
                retVal['max'] = var.max()
                retVal['mean'] = var.mean()
            else:
                retVal['min'] = 'not a numeric object'
                retVal['max'] = 'not a numeric object'
                retVal['mean'] = 'not a numeric object'
        retVal['shape'] = var.shape
        retVal['dtype'] = var.dtype
        retVal['size'] = var.size
        retVal['[0:{0}]'.format(
            len(var) - 1)] = list(var[0:MAX_ITEMS_TO_HANDLE])
        return retVal


class NdArrayItemsContainer:

    """Class to store ndarray items"""

    pass


#
# Resolver for Django Multi Value Dictionaries
#
class MultiValueDictResolver(DictResolver):

    """Class used to resolve from Django multi value dictionaries"""

    def resolve(self, var, attribute):
        """Provides an attribute from a variable"""
        if attribute in ('___len___', TOO_LARGE_ATTRIBUTE):
            return None

        if "(ID:" not in attribute:
            try:
                return var[attribute]
            except Exception:
                return getattr(var, attribute, None)

        expectedID = int(attribute.split("(ID:")[-1][:-1])
        for key in var.keys():
            if id(key) == expectedID:
                value = var.getlist(key)
                return value
        return None

    def getDictionary(self, var):
        """Provides the attributes of a variable as a dictionary"""
        retVal = {}
        count = 0
        for key in var.keys():
            count += 1
            value = var.getlist(key)
            key = "{0} (ID:{1})".format(self.keyToStr(key), id(key))
            retVal[key] = value
            if count > MAX_ITEMS_TO_HANDLE:
                retVal[TOO_LARGE_ATTRIBUTE] = TOO_LARGE_MESSAGE
                break

        retVal["___len___"] = len(var)
        return retVal


#
# Resolver for array.array
#
class ArrayResolver(BaseResolver):

    """Class used to resolve from array.array including some meta data"""

    TypeCodeMap = {
        "b": "int (signed char)",
        "B": "int (unsigned char)",
        "u": "Unicode character (Py_UNICODE)",
        "h": "int (signed short)",
        "H": "int (unsigned short)",
        "i": "int (signed int)",
        "I": "int (unsigned int)",
        "l": "int (signed long)",
        "L": "int (unsigned long)",
        "q": "int (signed long long)",
        "Q": "int (unsigned long long)",
        "f": "float (float)",
        "d": "float (double)",
    }

    def resolve(self, var, attribute):
        """Provides an attribute from a variable"""
        if attribute == 'itemsize':
            return var.itemsize

        if attribute == 'typecode':
            return var.typecode

        if attribute == 'type':
            if var.typecode in ArrayResolver.TypeCodeMap:
                return ArrayResolver.TypeCodeMap[var.typecode]
            return 'illegal type'

        if attribute.startswith('['):
            container = ArrayItemsContainer()
            count = 0
            for element in var:
                setattr(container, str(count), element)
                count += 1
                if count > MAX_ITEMS_TO_HANDLE:
                    setattr(container, TOO_LARGE_ATTRIBUTE, TOO_LARGE_MESSAGE)
                    break
            return container
        return None

    def getDictionary(self, var):
        """Provides the attributes of a variable as a dictionary"""
        retVal = {}
        retVal['typecode'] = var.typecode
        if var.typecode in ArrayResolver.TypeCodeMap:
            retVal['type'] = ArrayResolver.TypeCodeMap[var.typecode]
        else:
            retVal['type'] = 'illegal type'
        retVal['itemsize'] = var.itemsize
        retVal['[0:{0}]'.format(
            len(var) - 1)] = var.tolist()[0:MAX_ITEMS_TO_HANDLE]
        return retVal


class ArrayItemsContainer:

    """Class to store array.array items"""

    pass


defaultResolver = DefaultResolver()
dictResolver = DictResolver()
listResolver = ListResolver()
setResolver = SetResolver()
ndarrayResolver = NdArrayResolver()
multiValueDictResolver = MultiValueDictResolver()
arrayResolver = ArrayResolver()


#
# Methods to determine the type of a variable and the
# resolver class to use
#

_TypeMap = None


def _initTypeMap():
    """Initializes the type map"""
    global _TypeMap

    _TypeMap = [
        (type(None), None),
        (int, None),
        (float, None),
        (complex, None),
        (str, None),
        (tuple, listResolver),
        (list, listResolver),
        (dict, dictResolver)]

    try:
        _TypeMap.append((long, None))               # pylint: disable=E0602
    except Exception:
        pass    # not available on all python versions

    try:
        _TypeMap.append((unicode, None))            # pylint: disable=E0602
    except Exception:
        pass    # not available on all python versions

    try:
        _TypeMap.append((set, setResolver))         # pylint: disable=E0602
    except Exception:
        pass    # not available on all python versions

    try:
        _TypeMap.append((frozenset, setResolver))   # pylint: disable=E0602
    except Exception:
        pass    # not available on all python versions

    try:
        import array
        _TypeMap.append((array.array, arrayResolver))
    except ImportError:
        pass  # array.array may not be available

    try:
        import numpy
        _TypeMap.append((numpy.ndarray, ndarrayResolver))
    except ImportError:
        pass  # numpy may not be installed

    try:
        from django.utils.datastructures import MultiValueDict
        _TypeMap.insert(0, (MultiValueDict, multiValueDictResolver))
        # it should go before dict
    except ImportError:
        pass  # django may not be installed


def getType(obj):
    """Provides the type information for an object"""
    typeObject = type(obj)
    typeName = typeObject.__name__
    typeStr = str(typeObject)[8:-2]

    if typeStr.startswith(("PyQt5.", "PyQt4.")):
        resolver = None
    else:
        if _TypeMap is None:
            _initTypeMap()

        for typeData in _TypeMap:
            if isinstance(obj, typeData[0]):
                resolver = typeData[1]
                break
        else:
            resolver = defaultResolver
    return typeObject, typeName, typeStr, resolver



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2012  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Debugger break and watch point viewer"""

from ui.qt import Qt, QVBoxLayout, QWidget, QSplitter
from .bpointviewer import BreakPointViewer
from .wpointviewer import WatchPointViewer


class DebuggerBreakWatchPoints(QWidget):

    """Implements the debugger break and watch point viewer"""

    def __init__(self, parent, debugger):
        QWidget.__init__(self, parent)

        self.__debugger = debugger
        self.__createLayout()

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(1, 1, 1, 1)

        self.splitter = QSplitter(Qt.Vertical)
        self.breakPointViewer = BreakPointViewer(
            self.splitter, self.__debugger.getBreakPointModel())
        self.__watchPointViewer = WatchPointViewer(
            self.splitter, self.__debugger.getWatchPointModel())
        # TODO: temporary
        self.__watchPointViewer.setVisible(False)

        self.splitter.addWidget(self.breakPointViewer)
        self.splitter.addWidget(self.__watchPointViewer)

        self.splitter.setCollapsible(0, False)
        self.splitter.setCollapsible(1, False)

        verticalLayout.addWidget(self.splitter)

    def clear(self):
        """Clears everything"""
        self.breakPointViewer.clear()
        self.__watchPointViewer.clear()

    def setFocus(self):
        """Sets the focus to the break points window"""
        self.breakPointViewer.setFocus()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Codimension plugin manager"""

import logging
import os.path
import sys
from ui.qt import QObject, pyqtSignal
from yapsy.PluginManager import PluginManager
from utils.settings import SETTINGS_DIR, Settings
from distutils.version import StrictVersion


# List of the supported plugin categories, i.e. base class names
CATEGORIES = ["VersionControlSystemInterface",
              "WizardInterface"]

VENV_PLUGINS_PATH = os.path.normpath(
    os.path.dirname(sys.argv[0]) + '/../cdmplugins')


class CDMPluginManager(PluginManager, QObject):

    """Implements the codimension plugin manager"""

    sigPluginActivated = pyqtSignal(object)
    sigPluginDeactivated = pyqtSignal(object)

    NO_CONFLICT = 0
    # Same name plugin in system and user locations
    SYSTEM_USER_CONFLICT = 1
    # Plugin required incompatible version
    INCOMPATIBLE_IDE_VERSION_CONFLICT = 2
    # Newer version of the same name plugin
    VERSION_CONFLICT = 3
    # Does not derive from any of the supported interface
    BAD_BASE_CLASS = 4
    # The plugin raised exception during activation
    BAD_ACTIVATION = 5
    # Exception on basic methods
    BAD_INTERFACE = 6
    USER_DISABLED = 7

    def __init__(self):
        QObject.__init__(self)
        PluginManager.__init__(self, None,
                               [SETTINGS_DIR + "plugins",
                                "/usr/share/codimension3-plugins",
                                VENV_PLUGINS_PATH], "cdmp")

        self.inactivePlugins = {}   # Categorized inactive plugins
        self.activePlugins = {}     # Categorized active plugins
        self.unknownPlugins = []    # Unknown plugins

    def load(self):
        """Loads the found plugins"""
        # Now, let's check the plugins. They must be of known category.
        collectedPlugins = self.__collect()
        self.__applyDisabledPlugins(collectedPlugins)

        self.__checkIDECompatibility(collectedPlugins)
        self.__sysVsUserConflicts(collectedPlugins)
        self.__categoryConflicts(collectedPlugins)
        self.__activatePlugins(collectedPlugins)

        self.saveDisabledPlugins()

    def __collect(self):
        """Checks that the plugins belong to what is known"""
        self.collectPlugins()

        collectedPlugins = {}
        for plugin in self.getAllPlugins():
            recognised = False
            baseClasses = getBaseClassNames(plugin.plugin_object)
            for category in CATEGORIES:
                if category in baseClasses:
                    # OK, this plugin base has been recognised
                    recognised = True
                    newPlugin = CDMPluginInfo(plugin)
                    newPlugin.categoryName = category
                    if category in collectedPlugins:
                        collectedPlugins[category].append(newPlugin)
                    else:
                        collectedPlugins[category] = [newPlugin]
                    break

            if not recognised:
                logging.warning("Plugin of an unknown category is found at: " +
                                plugin.path + ". The plugin is disabled.")
                newPlugin = CDMPluginInfo(plugin)
                newPlugin.conflictType = CDMPluginManager.BAD_BASE_CLASS
                newPlugin.conflictMessage = "The plugin does not derive any " \
                                            "known plugin category interface"
                self.unknownPlugins.append(newPlugin)

        return collectedPlugins

    def __activatePlugins(self, collectedPlugins):
        """Activating the plugins"""
        from utils.globals import GlobalData

        for category in collectedPlugins:
            for plugin in collectedPlugins[category]:
                try:
                    plugin.getObject().activate(Settings(), GlobalData())
                    if category in self.activePlugins:
                        self.activePlugins[category].append(plugin)
                    else:
                        self.activePlugins[category] = [plugin]
                    self.sendPluginActivated(plugin)
                except Exception as excpt:
                    logging.error("Error activating plugin at " +
                                  plugin.getPath() +
                                  ". The plugin disabled. Error message: \n" +
                                  str(excpt))
                    plugin.conflictType = CDMPluginManager.BAD_ACTIVATION
                    plugin.conflictMessage = "Error activating the plugin"
                    if category in self.inactivePlugins:
                        self.inactivePlugins[category].append(plugin)
                    else:
                        self.inactivePlugins[category] = [plugin]

    def __checkIDECompatibility(self, collectedPlugins):
        """Checks that the plugins can be used with the current IDE"""
        from utils.globals import GlobalData

        toBeRemoved = []
        for category in collectedPlugins:
            for plugin in collectedPlugins[category]:
                try:
                    ideVer = GlobalData().version
                    if not plugin.getObject().isIDEVersionCompatible(ideVer):
                        # The plugin is incompatible. Disable it
                        logging.warning("The IDE version does not meet the " +
                                        plugin.getName() + " plugin "
                                        "requirements. The plugin is disabled."
                                        " (plugin path: " + plugin.getPath() +
                                        ")")
                        plugin.conflictType = \
                            CDMPluginManager.INCOMPATIBLE_IDE_VERSION_CONFLICT
                        plugin.conflictMessage = "The IDE version does not " \
                            "meet the plugin requirements."
                        self.unknownPlugins.append(plugin)
                        toBeRemoved.append(plugin.getPath())
                except Exception as excpt:
                    # Could not successfully call the interface method
                    logging.error("Error checking IDE version compatibility "
                                  "of plugin at " + plugin.getPath() +
                                  ". The plugin disabled. Error message: \n" +
                                  str(excpt))
                    plugin.conflictType = CDMPluginManager.BAD_INTERFACE
                    plugin.conflictMessage = "Error checking IDE version " \
                                             "compatibility"
                    if category in self.inactivePlugins:
                        self.inactivePlugins[category].append(plugin)
                    else:
                        self.inactivePlugins[category] = [plugin]
                    toBeRemoved.append(plugin.getPath())

        for path in toBeRemoved:
            for category in collectedPlugins:
                for plugin in collectedPlugins[category]:
                    if plugin.getPath() == path:
                        collectedPlugins[category].remove(plugin)
                        break

    def __applyDisabledPlugins(self, collectedPlugins):
        """Marks the disabled plugins in accordance to settings"""
        for disabledPlugin in Settings()['disabledPlugins']:
            # Parse the record
            try:
                conflictType, path, conflictMessage = \
                    CDMPluginInfo.parseDisabledLine(disabledPlugin)
            except Exception as excpt:
                logging.warning(str(excpt))
                continue

            found = False
            for category in collectedPlugins:
                for plugin in collectedPlugins[category]:
                    if plugin.getPath() == path:
                        found = True
                        plugin.conflictType = conflictType
                        plugin.conflictMessage = conflictMessage
                        if category in self.inactivePlugins:
                            self.inactivePlugins[category].append(plugin)
                        else:
                            self.inactivePlugins[category] = [plugin]
                        collectedPlugins[category].remove(plugin)
                        break
                if found:
                    break

            if not found:
                # Second try - search through the unknown plugins
                for plugin in self.unknownPlugins:
                    if plugin.getPath() == path:
                        found = True
                        plugin.conflictType = conflictType
                        plugin.conflictMessage = conflictMessage
                        break

            if not found:
                logging.warning("The disabled plugin at " + path +
                                " has not been found. The information that"
                                " the plugin is disabled will be deleted.")

    def __sysVsUserConflicts(self, collectedPlugins):
        """Checks for the system vs user plugin conflicts"""
        for category in collectedPlugins:
            self.__sysVsUserCategoryConflicts(category,
                                              collectedPlugins[category])

    def __sysVsUserCategoryConflicts(self, category, plugins):
        """Checks for the system vs user conflicts within one category"""
        def findIndexesByName(plugins, name):
            """Provides the plugin index by name"""
            result = []
            for index in range(len(plugins)):
                if plugins[index].getName() == name:
                    result.append(index)
            return result

        def hasUserPlugin(plugins, indexes):
            """True if has user plugins"""
            for index in indexes:
                if plugins[index].isUser():
                    return True
            return False

        index = 0
        while index < len(plugins):
            name = plugins[index].getName()
            sameNamePluginIndexes = findIndexesByName(plugins, name)
            if hasUserPlugin(plugins, sameNamePluginIndexes):
                # There is at least one user plugin
                # Disable all system plugins
                sameNamePluginIndexes.reverse()
                for checkIndex in sameNamePluginIndexes:
                    if not plugins[checkIndex].isUser():
                        logging.warning("The system wide plugin '" + name +
                                        "' at " +
                                        plugins[checkIndex].getPath() +
                                        " conflicts with a user plugin with "
                                        "the same name. The system wide "
                                        "plugin is automatically disabled.")
                        plugins[checkIndex].conflictType = \
                            CDMPluginManager.SYSTEM_USER_CONFLICT
                        plugins[checkIndex].conflictMessage = \
                            "It conflicts with a user plugin of the same name"
                        if category in self.inactivePlugins:
                            self.inactivePlugins[category].append(
                                plugins[checkIndex])
                        else:
                            self.inactivePlugins[category] = \
                                [plugins[checkIndex]]
                        del plugins[checkIndex]
                if plugins[index].getName() == name:
                    index += 1
            else:
                index += 1

    def __categoryConflicts(self, collectedPlugins):
        """Checks for version conflicts within the category"""
        for category in collectedPlugins:
            self.__singleCategoryConflicts(category,
                                           collectedPlugins[category])

    def __singleCategoryConflicts(self, category, plugins):
        """Checks a single category for name conflicts"""
        def findIndexesByName(plugins, name):
            """Provides the plugin index by name"""
            result = []
            for index in range(len(plugins)):
                if plugins[index].getName() == name:
                    result.append(index)
            return result

        index = 0
        while index < len(plugins):
            name = plugins[index].getName()
            sameNamePluginIndexes = findIndexesByName(plugins, name)
            if len(sameNamePluginIndexes) == 1:
                # The only plugin of the type. Keep it.
                index += 1
            else:
                # There are many. Check the versions and decide which to remove
                self.__resolveConflictByVersion(category, plugins,
                                                sameNamePluginIndexes)

    def __resolveConflictByVersion(self, category, plugins, indexes):
        """Resolves a single version conflict"""
        indexVersion = []
        for index in indexes:
            indexVersion.append((index,
                                 StrictVersion(plugins[index].getVersion())))

        # Sort basing on version
        indexVersion.sort(key=lambda indexVer: indexVer[1])

        # Disable everything except the last
        highVersion = indexVersion[-1][1]
        toBeDisabled = []
        for index in range(len(indexVersion) - 1):
            pluginIndex = indexVersion[index][0]
            logging.warning("The plugin '" + plugins[pluginIndex].getName() +
                            "' v." + plugins[pluginIndex].getVersion() +
                            " at " +
                            os.path.normpath(plugins[pluginIndex].getPath()) +
                            " conflicts with another plugin of the same name "
                            "and version " + str(highVersion) +
                            ". The former is disabled automatically.")
            toBeDisabled.append(pluginIndex)

        # Move the disabled to the inactive list
        toBeDisabled.sort()
        toBeDisabled.reverse()
        for index in toBeDisabled:
            plugins[index].conflictType = CDMPluginManager.VERSION_CONFLICT
            plugins[index].conflictMessage = \
                "It conflicts with another plugin of the same name"
            if category in self.inactivePlugins:
                self.inactivePlugins[category].append(plugins[index])
            else:
                self.inactivePlugins[category] = [plugins[index]]
            del plugins[index]

    def saveDisabledPlugins(self):
        """Saves the disabled plugins info into the settings"""
        value = []
        for category in self.inactivePlugins:
            for plugin in self.inactivePlugins[category]:
                line = plugin.getDisabledLine()
                if line is not None:
                    value.append(line)
        for plugin in self.unknownPlugins:
            line = plugin.getDisabledLine()
            if line is not None:
                value.append(line)
        Settings()['disabledPlugins'] = value

    def checkConflict(self, cdmPlugin):
        """Checks for the conflict and returns a message if so.

        If there is no conflict then returns None
        """
        # First, check the base class
        baseClasses = getBaseClassNames(cdmPlugin.getObject())
        category = None
        for registeredCategory in CATEGORIES:
            if registeredCategory in baseClasses:
                category = registeredCategory
                break
        if category is None:
            return "Plugin category is not recognised"

        # Second, IDE version compatibility
        from utils.globals import GlobalData
        try:
            ideVer = GlobalData().version
            if not cdmPlugin.getObject().isIDEVersionCompatible(ideVer):
                return "Plugin requires the other IDE version"
        except:
            # Could not successfully call the interface method
            return "Error checking IDE version compatibility"

        # Third, the other plugin with the same name is active
        if category in self.activePlugins:
            for plugin in self.activePlugins[category]:
                if plugin.getName() == cdmPlugin.getName():
                    return "Another plugin of the same name is active"

        return None

    def sendPluginActivated(self, plugin):
        """Emits the signal with the corresponding plugin"""
        self.sigPluginActivated.emit(plugin)
        plugin.getObject().pluginLogMessage.connect(self.__onPluginLogMessage)

    def sendPluginDeactivated(self, plugin):
        """Emits the signal with the corresponding plugin"""
        plugin.getObject().pluginLogMessage.disconnect(
            self.__onPluginLogMessage)
        self.sigPluginDeactivated.emit(plugin)

    @staticmethod
    def __onPluginLogMessage(logLevel, message):
        """Triggered when a plugin message is received"""
        logging.log(logLevel, str(message))


class CDMPluginInfo:

    """Holds info about a single plugin"""

    def __init__(self, pluginInfo):
        """The pluginInfo comes from yapsy"""
        # yapsy.PluginInfo
        self.__info = pluginInfo
        self.__isUser = self.__isUserPlugin()
        self.isEnabled = False
        # See CDMPluginManager constants
        self.conflictType = CDMPluginManager.NO_CONFLICT
        # One line message for UI/log
        self.conflictMessage = ""
        self.categoryName = None

    def isUser(self):
        """True if it is a user plugin"""
        return self.__isUser

    def __isUserPlugin(self):
        """True if it is a user plugin"""
        return self.getPath().startswith(SETTINGS_DIR)

    def getDisabledLine(self):
        """Used for the setting file"""
        if self.isEnabled is None or self.isEnabled:
            return None
        return str(self.conflictType) + ":::" + \
            self.__info.path + ":::" + \
            self.conflictMessage

    @staticmethod
    def parseDisabledLine(configLine):
        """Parser the config line and returns a tuple"""
        parts = configLine.split(":::", 2)
        if len(parts) != 3:
            raise ValueError("Incorrect disabled plugin description: " +
                             configLine)
        # (conflictType, path, conflictMessage)
        return (int(parts[0]), parts[1], parts[2])

    def getObject(self):
        """Provides a reference to the plugin object"""
        return self.__info.plugin_object

    def getPath(self):
        """Provides the plugin path"""
        return str(self.__info.path)

    def getName(self):
        """Provides the plugin name"""
        return str(self.__info.name)

    def getVersion(self):
        """Provides the plugin version"""
        return self.__info.details.get("Documentation", "Version")

    def getAuthor(self):
        """Provides the author name"""
        return self.__info.details.get("Documentation", "Author")

    def getDescription(self):
        """Provides the description"""
        return self.__info.details.get("Documentation", "Description")

    def getWebsite(self):
        """Provides the website"""
        return self.__info.details.get("Documentation", "Website")

    def getCopyright(self):
        """Provides the copyright"""
        return self.__info.details.get("Documentation", "Copyright")

    def getDetails(self):
        """Provides additional values from from the description section"""
        result = {}
        for name, value in self.__info.details.items("Documentation"):
            if name.lower() in ["version", "author", "description",
                                "website", "copyright"]:
                continue
            result[name] = value
        return result

    def disable(self, conflictType=CDMPluginManager.USER_DISABLED,
                conflictMessage=""):
        """Disables the plugin"""
        self.isEnabled = False
        self.conflictType = conflictType
        self.conflictMessage = conflictMessage

        if self.getObject().is_activated:
            if self.categoryName == "VersionControlSystemInterface":
                from utils.globals import GlobalData
                GlobalData().mainWindow.dismissVCSPlugin(self)
            self.getObject().deactivate()

    def enable(self):
        """Enables the plugin"""
        if not self.getObject().is_activated:
            from utils.globals import GlobalData
            self.getObject().activate(Settings(), GlobalData())

        self.isEnabled = True
        self.conflictType = CDMPluginManager.NO_CONFLICT
        self.conflictMessage = ""


def getBaseClassNames(inst):
    """Provides a list of base class names for the given instance"""
    baseNames = []

    def baseClassNames(inst, names):
        """Recursive retriever"""
        if hasattr(inst, "__bases__"):
            container = inst.__bases__
        else:
            container = inst.__class__.__bases__
        for base in container:
            names.append(base.__name__)
            if base.__name__ != "object":
                baseClassNames(base, names)
        return

    baseClassNames(inst, baseNames)
    return baseNames



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Run parameters dialog"""

import os
import os.path
import copy
from utils.runparams import RunParameters, RUN, PROFILE, DEBUG
from utils.run import parseCommandLineArguments, checkOutput
from .qt import (Qt, QDoubleValidator, QDialog, QDialogButtonBox, QVBoxLayout,
                 QSizePolicy, QLabel, QGridLayout, QHBoxLayout, QRadioButton,
                 QGroupBox, QPushButton, QFileDialog, QLineEdit, QTreeWidget,
                 QAbstractItemView, QTreeWidgetItem, QCheckBox)
from .itemdelegates import NoOutlineHeightDelegate


class EnvVarDialog(QDialog):

    """Single environment variable add/edit dialog"""

    def __init__(self, name='', value='', parent=None):
        QDialog.__init__(self, parent)

        self.name = name
        self.value = value

        self.__nameEdit = None
        self.__valueEdit = None
        self.__OKButton = None
        self.__createLayout()

        self.setWindowTitle("Environment variable")
        self.setMaximumHeight(self.sizeHint().height())
        self.setMaximumHeight(self.sizeHint().height())

        self.__nameEdit.setText(name)
        self.__valueEdit.setText(value)

        self.__nameEdit.setEnabled(name == "")
        self.__OKButton.setEnabled(name != "")

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(300, 50)
        self.setSizeGripEnabled(True)

        # Top level layout
        layout = QVBoxLayout(self)

        gridLayout = QGridLayout()
        nameLabel = QLabel("Name")
        gridLayout.addWidget(nameLabel, 0, 0)
        valueLabel = QLabel("Value")
        gridLayout.addWidget(valueLabel, 1, 0)
        self.__nameEdit = QLineEdit()
        self.__nameEdit.textChanged.connect(self.__nameChanged)
        gridLayout.addWidget(self.__nameEdit, 0, 1)
        self.__valueEdit = QLineEdit()
        self.__valueEdit.textChanged.connect(self.__valueChanged)
        gridLayout.addWidget(self.__valueEdit, 1, 1)
        layout.addLayout(gridLayout)

        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Ok |
                                     QDialogButtonBox.Cancel)
        self.__OKButton = buttonBox.button(QDialogButtonBox.Ok)
        self.__OKButton.setDefault(True)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.close)
        layout.addWidget(buttonBox)

    def __nameChanged(self, newName):
        """Triggered when a variable name is changed"""
        strippedName = str(newName).strip()
        self.__OKButton.setEnabled(strippedName != "" and
                                   ' ' not in strippedName)
        self.name = strippedName

    def __valueChanged(self, newValue):
        """Triggered when a variable value is changed"""
        self.value = newValue


class RunDialog(QDialog):

    """Run parameters dialog implementation"""

    ACTION_TO_VERB = {RUN: 'Run',
                      PROFILE: 'Profile',
                      DEBUG: 'Debug'}

    # See utils.run for runParameters
    def __init__(self, path, runParameters,
                 profilerParams, debuggerParams,
                 action, parent=None):
        QDialog.__init__(self, parent)

        # Used as a return value
        self.runParams = copy.deepcopy(runParameters)
        self.profilerParams = copy.deepcopy(profilerParams)
        self.debuggerParams = copy.deepcopy(debuggerParams)

        self.__action = action

        # Avoid pylint complains
        self.__argsEdit = None
        self.__scriptWDRButton = None
        self.__dirRButton = None
        self.__dirEdit = None
        self.__dirSelectButton = None
        self.__inheritParentRButton = None
        self.__inheritParentPlusRButton = None
        self.__inhPlusEnvTable = None
        self.__addInhButton = None
        self.__delInhButton = None
        self.__editInhButton = None
        self.__specificRButton = None
        self.__specEnvTable = None
        self.__addSpecButton = None
        self.__delSpecButton = None
        self.__editSpecButton = None
        self.__runButton = None
        self.__nodeLimitEdit = None
        self.__edgeLimitEdit = None
        self.__debugChildCheckBox = None
        self.__edgeLimitValidator = None
        self.__nodeLimitValidator = None
        self.__intSelectButton = None
        self.__intEdit = None
        self.__redirectedRButton = None
        self.__customIntRButton = None
        self.__customTermRButton = None
        self.__stopAtFirstCheckBox = None
        self.__traceInterpreterCheckBox = None
        self.__autoforkCheckBox = None
        self.__reportExceptionCheckBox = None
        self.__termEdit = None
        self.__inheritedInterpreterRButton = None

        self.__createLayout()
        self.setWindowTitle(RunDialog.ACTION_TO_VERB[action] +
                            ' parameters for ' + path)

        self.__populateValues()

    def __populateValues(self):
        """Populates the dialog UI controls"""
        self.__argsEdit.setText(self.runParams['arguments'])

        self.__populateWorkingDir()
        self.__populateEnvironment()
        self.__populateInterpreter()
        self.__populateIO()

        if self.__action == PROFILE:
            self.__populateProfile()
        elif self.__action == DEBUG:
            self.__populateDebug()

        self.__setRunButtonProps()

    def __populateWorkingDir(self):
        """Populates the working directory"""
        if self.runParams['useScriptLocation']:
            self.__scriptWDRButton.setChecked(True)
            self.__dirEdit.setEnabled(False)
            self.__dirSelectButton.setEnabled(False)
        else:
            self.__dirRButton.setChecked(True)
            self.__dirEdit.setEnabled(True)
            self.__dirSelectButton.setEnabled(True)
        self.__dirEdit.setText(self.runParams['specificDir'])

    def __populateEnvironment(self):
        """Populates the environment variables"""
        self.__populateTable(self.__inhPlusEnvTable,
                             self.runParams['additionToParentEnv'])
        self.__populateTable(self.__specEnvTable,
                             self.runParams['specificEnv'])

        if self.runParams['envType'] == RunParameters.InheritParentEnv:
            self.__inheritParentRButton.setChecked(True)
            self.__setEnabledInheritedPlusEnv(False)
            self.__setEnabledSpecificEnv(False)
        elif self.runParams['envType'] == RunParameters.InheritParentEnvPlus:
            self.__inheritParentPlusRButton.setChecked(True)
            self.__setEnabledSpecificEnv(False)
        else:
            self.__specificRButton.setChecked(True)
            self.__setEnabledInheritedPlusEnv(False)

    def __populateInterpreter(self):
        """Populates the interpreter"""
        if self.runParams['useInherited']:
            self.__inheritedInterpreterRButton.setChecked(True)
            self.__intEdit.setEnabled(False)
            self.__intSelectButton.setEnabled(False)
        else:
            self.__customIntRButton.setChecked(True)
            self.__intEdit.setEnabled(True)
            self.__intSelectButton.setEnabled(True)
        self.__intEdit.setText(self.runParams['customInterpreter'])

    def __populateIO(self):
        """Populate I/O"""
        if self.runParams['redirected']:
            self.__redirectedRButton.setChecked(True)
            self.__termEdit.setEnabled(False)
        else:
            self.__customTermRButton.setChecked(True)
            self.__termEdit.setEnabled(True)
        self.__termEdit.setText(self.runParams['customTerminal'])
        self.__termEdit.setToolTip(
            'Use ${prog} substitution if needed.\n'
            'Otherwise the command line is attached at the end.\n'
            'E.g.: xterm -e /bin/bash -c "${prog}; /bin/bash" &')

    def __populateProfile(self):
        """Populates profile"""
        if self.profilerParams.nodeLimit < 0.0 or \
           self.profilerParams.nodeLimit > 100.0:
            self.profilerParams.nodeLimit = 1.0
        self.__nodeLimitEdit.setText(str(self.profilerParams.nodeLimit))
        if self.profilerParams.edgeLimit < 0.0 or \
           self.profilerParams.edgeLimit > 100.0:
            self.profilerParams.edgeLimit = 1.0
        self.__edgeLimitEdit.setText(str(self.profilerParams.edgeLimit))

    def __populateDebug(self):
        """Populates debug"""
        self.__reportExceptionCheckBox.setChecked(
            self.debuggerParams.reportExceptions)
        self.__traceInterpreterCheckBox.setChecked(
            self.debuggerParams.traceInterpreter)
        self.__stopAtFirstCheckBox.setChecked(
            self.debuggerParams.stopAtFirstLine)
        self.__autoforkCheckBox.setChecked(self.debuggerParams.autofork)
        self.__debugChildCheckBox.setChecked(self.debuggerParams.followChild)
        self.__debugChildCheckBox.setEnabled(self.debuggerParams.autofork)

    @staticmethod
    def __populateTable(table, dictionary):
        """Populates the given table"""
        for key, value in dictionary.items():
            item = QTreeWidgetItem([key, value])
            table.addTopLevelItem(item)
        if dictionary:
            table.setCurrentItem(table.topLevelItem(0))

    def __setEnabledInheritedPlusEnv(self, value):
        """Disables/enables 'inherited and add' section controls"""
        self.__inhPlusEnvTable.setEnabled(value)
        self.__addInhButton.setEnabled(value)
        self.__delInhButton.setEnabled(value)
        self.__editInhButton.setEnabled(value)

    def __setEnabledSpecificEnv(self, value):
        """Disables/enables 'specific env' section controls"""
        self.__specEnvTable.setEnabled(value)
        self.__addSpecButton.setEnabled(value)
        self.__delSpecButton.setEnabled(value)
        self.__editSpecButton.setEnabled(value)

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(650, 300)
        self.setSizeGripEnabled(True)

        layout = QVBoxLayout(self)  # top level layout
        layout.addLayout(self.__getArgLayout())
        layout.addWidget(self.__getWorkingDirGroupbox())
        layout.addWidget(self.__getEnvGroupbox())
        layout.addWidget(self.__getInterpreterGroupbox())
        layout.addWidget(self.__getIOGroupbox())

        if self.__action == PROFILE:
            layout.addWidget(self.__getProfileLimitsGroupbox())
        elif self.__action == DEBUG:
            layout.addWidget(self.__getDebugGroupbox())

        # Buttons at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Cancel)
        self.__runButton = buttonBox.addButton(
            RunDialog.ACTION_TO_VERB[self.__action],
            QDialogButtonBox.AcceptRole)
        self.__runButton.setDefault(True)
        self.__runButton.clicked.connect(self.onAccept)
        layout.addWidget(buttonBox)

        buttonBox.rejected.connect(self.close)

    def __getArgLayout(self):
        """Provides the arguments layout"""
        argsLabel = QLabel("Command line arguments")
        self.__argsEdit = QLineEdit()
        self.__argsEdit.textChanged.connect(self.__argsChanged)
        argsLayout = QHBoxLayout()
        argsLayout.addWidget(argsLabel)
        argsLayout.addWidget(self.__argsEdit)
        return argsLayout

    @staticmethod
    def __getSizePolicy(item):
        """Provides a common size policy"""
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(item.sizePolicy().hasHeightForWidth())
        return sizePolicy

    def __getWorkingDirGroupbox(self):
        """Provides the working dir groupbox"""
        workDirGroupbox = QGroupBox('Working Directory', self)
        workDirGroupbox.setSizePolicy(self.__getSizePolicy(workDirGroupbox))

        gridLayoutWD = QGridLayout(workDirGroupbox)
        self.__scriptWDRButton = QRadioButton("&Use script location",
                                              workDirGroupbox)
        gridLayoutWD.addWidget(self.__scriptWDRButton, 0, 0)
        self.__scriptWDRButton.clicked.connect(lambda: self.__wdDir(True))

        self.__dirRButton = QRadioButton("Select &directory", workDirGroupbox)
        gridLayoutWD.addWidget(self.__dirRButton, 1, 0)
        self.__dirRButton.clicked.connect(lambda: self.__wdDir(False))

        self.__dirEdit = QLineEdit(workDirGroupbox)
        gridLayoutWD.addWidget(self.__dirEdit, 1, 1)
        self.__dirEdit.textChanged.connect(self.__workingDirChanged)

        self.__dirSelectButton = QPushButton("...", workDirGroupbox)
        gridLayoutWD.addWidget(self.__dirSelectButton, 1, 2)
        self.__dirSelectButton.clicked.connect(self.__selectDirClicked)
        return workDirGroupbox

    def __getEnvGroupbox(self):
        """Provides the environment groupbox"""
        envGroupbox = QGroupBox('Environment', self)
        envGroupbox.setSizePolicy(self.__getSizePolicy(envGroupbox))

        layoutEnv = QVBoxLayout(envGroupbox)
        self.__inheritParentRButton = QRadioButton("Inherit &parent",
                                                   envGroupbox)
        self.__inheritParentRButton.clicked.connect(self.__inhClicked)
        layoutEnv.addWidget(self.__inheritParentRButton)

        self.__inheritParentPlusRButton = QRadioButton(
            "Inherit parent and add/&modify", envGroupbox)
        self.__inheritParentPlusRButton.clicked.connect(self.__inhPlusClicked)
        layoutEnv.addWidget(self.__inheritParentPlusRButton)
        hInhPlusLayout = QHBoxLayout()
        self.__inhPlusEnvTable = QTreeWidget()
        self.__inhPlusEnvTable.itemActivated.connect(
            lambda x, y: self.__editInhClicked())
        self.__tuneTable(self.__inhPlusEnvTable)
        hInhPlusLayout.addWidget(self.__inhPlusEnvTable)
        vInhPlusLayout = QVBoxLayout()
        self.__addInhButton = QPushButton('Add')
        self.__addInhButton.clicked.connect(self.__addInhClicked)
        vInhPlusLayout.addWidget(self.__addInhButton)
        self.__delInhButton = QPushButton('Delete')
        self.__delInhButton.clicked.connect(self.__delInhClicked)
        vInhPlusLayout.addWidget(self.__delInhButton)
        self.__editInhButton = QPushButton("Edit")
        self.__editInhButton.clicked.connect(self.__editInhClicked)
        vInhPlusLayout.addWidget(self.__editInhButton)
        hInhPlusLayout.addLayout(vInhPlusLayout)
        layoutEnv.addLayout(hInhPlusLayout)

        self.__specificRButton = QRadioButton("&Specific", envGroupbox)
        self.__specificRButton.clicked.connect(self.__specClicked)
        layoutEnv.addWidget(self.__specificRButton)
        hSpecLayout = QHBoxLayout()
        self.__specEnvTable = QTreeWidget()
        self.__specEnvTable.itemActivated.connect(
            lambda x, y: self.__editSpecClicked())
        self.__tuneTable(self.__specEnvTable)
        hSpecLayout.addWidget(self.__specEnvTable)
        vSpecLayout = QVBoxLayout()
        self.__addSpecButton = QPushButton('Add')
        self.__addSpecButton.clicked.connect(self.__addSpecClicked)
        vSpecLayout.addWidget(self.__addSpecButton)
        self.__delSpecButton = QPushButton('Delete')
        self.__delSpecButton.clicked.connect(self.__delSpecClicked)
        vSpecLayout.addWidget(self.__delSpecButton)
        self.__editSpecButton = QPushButton("Edit")
        self.__editSpecButton.clicked.connect(self.__editSpecClicked)
        vSpecLayout.addWidget(self.__editSpecButton)
        hSpecLayout.addLayout(vSpecLayout)
        layoutEnv.addLayout(hSpecLayout)
        return envGroupbox

    def __getInterpreterGroupbox(self):
        """Creates the interpreter groupbox"""
        interpreterGroupbox = QGroupBox('Python Interpreter', self)
        interpreterGroupbox.setSizePolicy(
            self.__getSizePolicy(interpreterGroupbox))

        gridLayoutInt = QGridLayout(interpreterGroupbox)
        self.__inheritedInterpreterRButton = QRadioButton(
            "&Inherited", interpreterGroupbox)
        gridLayoutInt.addWidget(self.__inheritedInterpreterRButton, 0, 0)
        self.__inheritedInterpreterRButton.clicked.connect(
            lambda: self.__interpreter(True))

        self.__customIntRButton = QRadioButton(
            "Select interpreter (series &3)", interpreterGroupbox)
        gridLayoutInt.addWidget(self.__customIntRButton, 1, 0)
        self.__customIntRButton.clicked.connect(
            lambda: self.__interpreter(False))

        self.__intEdit = QLineEdit(interpreterGroupbox)
        gridLayoutInt.addWidget(self.__intEdit, 1, 1)
        self.__intEdit.textChanged.connect(self.__interpreterChanged)

        self.__intSelectButton = QPushButton("...", interpreterGroupbox)
        gridLayoutInt.addWidget(self.__intSelectButton, 1, 2)
        self.__intSelectButton.clicked.connect(self.__selectIntClicked)
        return interpreterGroupbox

    def __getIOGroupbox(self):
        """Creates the interpreter groupbox"""
        ioGroupbox = QGroupBox('Input/output', self)
        ioGroupbox.setSizePolicy(self.__getSizePolicy(ioGroupbox))

        gridLayoutInt = QGridLayout(ioGroupbox)
        self.__redirectedRButton = QRadioButton("&Redirected I/O", ioGroupbox)
        gridLayoutInt.addWidget(self.__redirectedRButton, 0, 0)
        self.__redirectedRButton.clicked.connect(
            lambda: self.__redirected(True))

        self.__customTermRButton = QRadioButton("Custom terminal string",
                                                ioGroupbox)
        gridLayoutInt.addWidget(self.__customTermRButton, 1, 0)
        self.__customTermRButton.clicked.connect(
            lambda: self.__redirected(False))

        self.__termEdit = QLineEdit(ioGroupbox)
        gridLayoutInt.addWidget(self.__termEdit, 1, 1)
        self.__termEdit.textChanged.connect(self.__customTermChanged)
        return ioGroupbox

    def __getProfileLimitsGroupbox(self):
        """Creates the profile limits groupbox"""
        limitsGroupbox = QGroupBox('Profiler diagram limits (IDE wide)', self)
        limitsGroupbox.setSizePolicy(self.__getSizePolicy(limitsGroupbox))

        layoutLimits = QGridLayout(limitsGroupbox)
        self.__nodeLimitEdit = QLineEdit()
        self.__nodeLimitEdit.textEdited.connect(self.__setRunButtonProps)
        self.__nodeLimitValidator = QDoubleValidator(0.0, 100.0, 2, self)
        self.__nodeLimitValidator.setNotation(
            QDoubleValidator.StandardNotation)
        self.__nodeLimitEdit.setValidator(self.__nodeLimitValidator)
        nodeLimitLabel = QLabel("Hide nodes below")
        self.__edgeLimitEdit = QLineEdit()
        self.__edgeLimitEdit.textEdited.connect(self.__setRunButtonProps)
        self.__edgeLimitValidator = QDoubleValidator(0.0, 100.0, 2, self)
        self.__edgeLimitValidator.setNotation(
            QDoubleValidator.StandardNotation)
        self.__edgeLimitEdit.setValidator(self.__edgeLimitValidator)
        edgeLimitLabel = QLabel("Hide edges below")
        layoutLimits.addWidget(nodeLimitLabel, 0, 0)
        layoutLimits.addWidget(self.__nodeLimitEdit, 0, 1)
        layoutLimits.addWidget(QLabel("%"), 0, 2)
        layoutLimits.addWidget(edgeLimitLabel, 1, 0)
        layoutLimits.addWidget(self.__edgeLimitEdit, 1, 1)
        layoutLimits.addWidget(QLabel("%"), 1, 2)
        return limitsGroupbox

    def __getDebugGroupbox(self):
        """Creates the debug settings groupbox"""
        dbgGroupbox = QGroupBox('Debugger (IDE wide)', self)
        dbgGroupbox.setSizePolicy(self.__getSizePolicy(dbgGroupbox))

        dbgLayout = QVBoxLayout(dbgGroupbox)
        self.__reportExceptionCheckBox = QCheckBox("Report &exceptions")
        self.__reportExceptionCheckBox.stateChanged.connect(
            self.__onReportExceptionChanged)
        self.__traceInterpreterCheckBox = QCheckBox("T&race interpreter libs")
        self.__traceInterpreterCheckBox.stateChanged.connect(
            self.__onTraceInterpreterChanged)
        self.__stopAtFirstCheckBox = QCheckBox("Stop at first &line")
        self.__stopAtFirstCheckBox.stateChanged.connect(
            self.__onStopAtFirstChanged)
        self.__autoforkCheckBox = QCheckBox("&Fork without asking")
        self.__autoforkCheckBox.stateChanged.connect(self.__onAutoforkChanged)
        self.__debugChildCheckBox = QCheckBox("Debu&g child process")
        self.__debugChildCheckBox.stateChanged.connect(self.__onDebugChild)

        dbgLayout.addWidget(self.__reportExceptionCheckBox)
        dbgLayout.addWidget(self.__traceInterpreterCheckBox)
        dbgLayout.addWidget(self.__stopAtFirstCheckBox)
        dbgLayout.addWidget(self.__autoforkCheckBox)
        dbgLayout.addWidget(self.__debugChildCheckBox)
        return dbgGroupbox

    @staticmethod
    def __tuneTable(table):
        """Sets the common settings for a table"""
        table.setAlternatingRowColors(True)
        table.setRootIsDecorated(False)
        table.setItemsExpandable(False)
        table.setUniformRowHeights(True)
        table.setSelectionMode(QAbstractItemView.SingleSelection)
        table.setSelectionBehavior(QAbstractItemView.SelectRows)
        table.setItemDelegate(NoOutlineHeightDelegate(4))
        table.setHeaderLabels(["Variable", "Value"])

        header = table.header()
        header.setSortIndicator(0, Qt.AscendingOrder)
        header.setSortIndicatorShown(True)
        header.setSectionsClickable(True)
        table.setSortingEnabled(True)

    def __wdDir(self, useScriptLocation):
        """Working dir radio selection changed"""
        self.__dirEdit.setEnabled(not useScriptLocation)
        self.__dirSelectButton.setEnabled(not useScriptLocation)
        self.runParams['useScriptLocation'] = useScriptLocation
        self.__setRunButtonProps()

    def __interpreter(self, useInherited):
        """Interpreter radio selection changed"""
        self.__intEdit.setEnabled(not useInherited)
        self.__intSelectButton.setEnabled(not useInherited)
        self.runParams['useInherited'] = useInherited
        self.__setRunButtonProps()

    def __redirected(self, redirected):
        """I/O radio button changed"""
        self.__termEdit.setEnabled(not redirected)
        self.runParams['redirected'] = redirected
        self.__setRunButtonProps()

    def __customTermChanged(self, value):
        """Triggered when a custom terminal string changed"""
        value = str(value).strip()
        self.runParams['customTerminal'] = value
        self.__setRunButtonProps()

    def __argsChanged(self, value):
        """Triggered when cmd line args are changed"""
        value = str(value).strip()
        self.runParams['arguments'] = value
        self.__setRunButtonProps()

    def __workingDirChanged(self, value):
        """Triggered when a working dir value is changed"""
        value = str(value)
        self.runParams['specificDir'] = value
        self.__setRunButtonProps()

    def __interpreterChanged(self, value):
        """Triggered when an interpreter is changed"""
        value = str(value).strip()
        self.runParams['customInterpreter'] = value
        self.__setRunButtonProps()

    def __onCloseChanged(self, state):
        """Triggered when the close terminal check box changed"""
        self.runParams['closeTerminal'] = state != 0

    def __onReportExceptionChanged(self, state):
        """Triggered when exception report check box changed"""
        self.debuggerParams.reportExceptions = state != 0

    def __onTraceInterpreterChanged(self, state):
        """Triggered when trace interpreter changed"""
        self.debuggerParams.traceInterpreter = state != 0

    def __onStopAtFirstChanged(self, state):
        """Triggered when stop at first changed"""
        self.debuggerParams.stopAtFirstLine = state != 0

    def __onAutoforkChanged(self, state):
        """Triggered when autofork changed"""
        self.debuggerParams.autofork = state != 0
        self.__debugChildCheckBox.setEnabled(self.debuggerParams.autofork)

    def __onDebugChild(self, state):
        """Triggered when debug child changed"""
        self.debuggerParams.followChild = state != 0

    def __argumentsOK(self):
        """Returns True if the arguments are OK"""
        try:
            parseCommandLineArguments(self.runParams['arguments'])
            return True
        except:
            return False

    def __dirOK(self):
        """Returns True if the working dir is OK"""
        if self.__scriptWDRButton.isChecked():
            return True
        return os.path.isdir(self.__dirEdit.text())

    def __interpreterOK(self):
        """Checks if the interpreter is OK"""
        if self.__inheritedInterpreterRButton.isChecked():
            return True
        path = self.__intEdit.text().strip()
        if not path:
            return 'No executable specified'
        try:
            code = "from __future__ import print_function; " \
                "import sys; print(sys.version_info.major)"
            output = checkOutput(path + ' -c "' + code + '"', useShell=True)
            output = output.strip()
            if output != '3':
                return 'Only python series 3 is supported ' \
                    '(provided: series ' + output + ')'
        except:
            return 'Error checking the provided interpreter'

    def __ioOK(self):
        """Checks if the IO is correct"""
        if self.__redirectedRButton.isChecked():
            return True

        term = self.__termEdit.text().strip()
        if not term:
            return 'No custom terminal line specified'

    def __setRunButtonProps(self, _=None):
        """Enable/disable run button and set its tooltip"""
        if not self.__argumentsOK():
            self.__runButton.setEnabled(False)
            self.__runButton.setToolTip("No closing quotation in arguments")
            return

        if not self.__dirOK():
            self.__runButton.setEnabled(False)
            self.__runButton.setToolTip("The given working "
                                        "dir is not found")
            return

        interpreterOK = self.__interpreterOK()
        if isinstance(interpreterOK, str):
            self.__runButton.setEnabled(False)
            self.__runButton.setToolTip('Invalid interpreter. ' +
                                        interpreterOK)
            return

        ioOK = self.__ioOK()
        if isinstance(ioOK, str):
            self.__runButton.setEnabled(False)
            self.__runButton.setToolTip('Invalid terminal. ' + ioOK)
            return

        if self.__nodeLimitEdit is not None:
            txt = self.__nodeLimitEdit.text().strip()
            try:
                value = float(txt)
                if value < 0.0 or value > 100.0:
                    raise Exception("Out of range")
            except:
                self.__runButton.setEnabled(False)
                self.__runButton.setToolTip("The given node limit "
                                            "is out of range")
                return

        if self.__edgeLimitEdit is not None:
            txt = self.__edgeLimitEdit.text().strip()
            try:
                value = float(txt)
                if value < 0.0 or value > 100.0:
                    raise Exception("Out of range")
            except:
                self.__runButton.setEnabled(False)
                self.__runButton.setToolTip("The given edge limit "
                                            "is out of range")
                return

        self.__runButton.setEnabled(True)
        self.__runButton.setToolTip(
            "Save parameters and " +
            RunDialog.ACTION_TO_VERB[self.__action].lower() + " script")

    def __selectDirClicked(self):
        """Selects the script working dir"""
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog | QFileDialog.ShowDirsOnly
        dirName = QFileDialog.getExistingDirectory(
            self, "Select the script working directory",
            self.__dirEdit.text(), options=options)

        if dirName:
            self.__dirEdit.setText(os.path.normpath(dirName))

    def __selectIntClicked(self):
        """Selects a python interpreter"""
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        path, _ = QFileDialog.getOpenFileName(
            self, "Select python series 3 interpreter",
            options=options)

        if path:
            self.__intEdit.setText(os.path.normpath(path))
        self.__setRunButtonProps()

    def __inhClicked(self):
        """Inerit parent env radio button clicked"""
        self.__setEnabledInheritedPlusEnv(False)
        self.__setEnabledSpecificEnv(False)
        self.runParams['envType'] = RunParameters.InheritParentEnv

    def __inhPlusClicked(self):
        """Inherit parent and add radio button clicked"""
        self.__setEnabledInheritedPlusEnv(True)
        self.__setEnabledSpecificEnv(False)
        self.runParams['envType'] = RunParameters.InheritParentEnvPlus

        if self.__inhPlusEnvTable.selectedIndexes():
            self.__delInhButton.setEnabled(True)
            self.__editInhButton.setEnabled(True)
        else:
            self.__delInhButton.setEnabled(False)
            self.__editInhButton.setEnabled(False)

    def __specClicked(self):
        """Specific env radio button clicked"""
        self.__setEnabledInheritedPlusEnv(False)
        self.__setEnabledSpecificEnv(True)
        self.runParams['envType'] = RunParameters.SpecificEnvironment

        if self.__specEnvTable.selectedIndexes():
            self.__delSpecButton.setEnabled(True)
            self.__editSpecButton.setEnabled(True)
        else:
            self.__delSpecButton.setEnabled(False)
            self.__editSpecButton.setEnabled(False)

    @staticmethod
    def __delAndInsert(table, name, value):
        """Deletes an item by name if so; insert new; highlight it"""
        for index in range(table.topLevelItemCount()):
            item = table.topLevelItem(index)
            if str(item.text(0)) == name:
                table.takeTopLevelItem(index)
                break

        item = QTreeWidgetItem([name, value])
        table.addTopLevelItem(item)
        table.setCurrentItem(item)
        return item

    def __addInhClicked(self):
        """Add env var button clicked"""
        dlg = EnvVarDialog()
        if dlg.exec_() == QDialog.Accepted:
            name = str(dlg.name)
            value = str(dlg.value)
            self.__delAndInsert(self.__inhPlusEnvTable, name, value)
            self.runParams['additionToParentEnv'][name] = value
            self.__delInhButton.setEnabled(True)
            self.__editInhButton.setEnabled(True)

    def __addSpecClicked(self):
        """Add env var button clicked"""
        dlg = EnvVarDialog()
        if dlg.exec_() == QDialog.Accepted:
            name = str(dlg.name)
            value = str(dlg.value)
            self.__delAndInsert(self.__specEnvTable, name, value)
            self.runParams['specificEnv'][name] = value
            self.__delSpecButton.setEnabled(True)
            self.__editSpecButton.setEnabled(True)

    def __delInhClicked(self):
        """Delete the highlighted variable"""
        if self.__inhPlusEnvTable.topLevelItemCount() == 0:
            return

        name = self.__inhPlusEnvTable.currentItem().text(0)
        for index in range(self.__inhPlusEnvTable.topLevelItemCount()):
            item = self.__inhPlusEnvTable.topLevelItem(index)
            if name == item.text(0):
                self.__inhPlusEnvTable.takeTopLevelItem(index)
                break

        del self.runParams['additionToParentEnv'][str(name)]
        if self.__inhPlusEnvTable.topLevelItemCount() == 0:
            self.__delInhButton.setEnabled(False)
            self.__editInhButton.setEnabled(False)
        else:
            self.__inhPlusEnvTable.setCurrentItem(
                self.__inhPlusEnvTable.topLevelItem(0))

    def __delSpecClicked(self):
        """Delete the highlighted variable"""
        if self.__specEnvTable.topLevelItemCount() == 0:
            return

        name = self.__specEnvTable.currentItem().text(0)
        for index in range(self.__specEnvTable.topLevelItemCount()):
            item = self.__specEnvTable.topLevelItem(index)
            if name == item.text(0):
                self.__specEnvTable.takeTopLevelItem(index)
                break

        del self.runParams['specificEnv'][str(name)]
        if self.__specEnvTable.topLevelItemCount() == 0:
            self.__delSpecButton.setEnabled(False)
            self.__editSpecButton.setEnabled(False)
        else:
            self.__specEnvTable.setCurrentItem(
                self.__specEnvTable.topLevelItem(0))

    def __editInhClicked(self):
        """Edits the highlighted variable"""
        if self.__inhPlusEnvTable.topLevelItemCount() == 0:
            return

        item = self.__inhPlusEnvTable.currentItem()
        dlg = EnvVarDialog(str(item.text(0)), str(item.text(1)), self)
        if dlg.exec_() == QDialog.Accepted:
            name = str(dlg.name)
            value = str(dlg.value)
            self.__delAndInsert(self.__inhPlusEnvTable, name, value)
            self.runParams['additionToParentEnv'][name] = value

    def __editSpecClicked(self):
        """Edits the highlighted variable"""
        if self.__specEnvTable.topLevelItemCount() == 0:
            return

        item = self.__specEnvTable.currentItem()
        dlg = EnvVarDialog(str(item.text(0)), str(item.text(1)), self)
        if dlg.exec_() == QDialog.Accepted:
            name = str(dlg.name)
            value = str(dlg.value)
            self.__delAndInsert(self.__specEnvTable, name, value)
            self.runParams['specificEnv'][name] = value

    def onAccept(self):
        """Saves the selected terminal and profiling values"""
        if self.__action == PROFILE:
            self.profilerParams.nodeLimit = float(
                self.__nodeLimitEdit.text())
            self.profilerParams.edgeLimit = float(
                self.__edgeLimitEdit.text())

        self.accept()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Relay for values stored on the disk"""

# Some values could be stored for a project or for an IDE. This module
# provides a set of the functions to relay a call to the appropriate
# location

from os.path import relpath, dirname
from .globals import GlobalData
from .settings import Settings


def getCollapsedGroups(fileName):
    """Provides None if not found"""
    project = GlobalData().project
    if project.isLoaded():
        if project.isProjectFile(fileName):
            key = relpath(fileName, dirname(project.fileName))
            return project.getFileGroups(key)
    return Settings().getFileGroups(fileName)


def setCollapsedGroups(fileName, groups):
    """Sets the file collapsed groups"""
    project = GlobalData().project
    if project.isLoaded():
        if project.isProjectFile(fileName):
            key = relpath(fileName, dirname(project.fileName))
            project.setFileGroups(key, groups)
            return
    Settings().setFileGroups(fileName, groups)


def addCollapsedGroup(fileName, group):
    """Adds a group into a list collapsed group for a file"""
    project = GlobalData().project
    if project.isLoaded():
        if project.isProjectFile(fileName):
            key = relpath(fileName, dirname(project.fileName))
            project.addFileGroup(key, group)
            return
    Settings().addFileGroup(fileName, group)


def removeCollapsedGroup(fileName, group):
    """Removes a group from a list of collapsed groups for a file"""
    project = GlobalData().project
    if project.isLoaded():
        if project.isProjectFile(fileName):
            key = relpath(fileName, dirname(project.fileName))
            project.removeFileGroup(key, group)
            return
    Settings().removeFileGroup(fileName, group)


def getFileEncoding(fileName):
    """Provides None if not found"""
    project = GlobalData().project
    if project.isLoaded():
        if project.isProjectFile(fileName):
            key = relpath(fileName, dirname(project.fileName))
            return project.getFileEncoding(key)
    return Settings().getFileEncoding(fileName)


def setFileEncoding(fileName, encoding):
    """Sets the encoding for the file"""
    project = GlobalData().project
    if project.isLoaded():
        if project.isProjectFile(fileName):
            key = relpath(fileName, dirname(project.fileName))
            project.setFileEncoding(key, encoding)
            return
    Settings().setFileEncoding(fileName, encoding)


def getRunParameters(fileName):
    """Provides the run parameters"""
    project = GlobalData().project
    if project.isLoaded():
        if project.isProjectFile(fileName):
            key = relpath(fileName, dirname(project.fileName))
            return project.getRunParameters(key)
    return Settings().getRunParameters(fileName)


def addRunParams(fileName, params):
    """Registers new latest run parameters"""
    project = GlobalData().project
    if project.isLoaded():
        if project.isProjectFile(fileName):
            key = relpath(fileName, dirname(project.fileName))
            project.addRunParameters(key, params)
            return
    Settings().addRunParameters(fileName, params)


def getFilePosition(fileName):
    """Provides the position or (-1,-1,-1,-1,-1) if not found"""
    project = GlobalData().project
    if project.isLoaded():
        if project.isProjectFile(fileName):
            key = relpath(fileName, dirname(project.fileName))
            return project.getFilePosition(key)
    return Settings().getFilePosition(fileName)


def updateFilePosition(fileName, line, pos, firstLine,
                       horizontalPos, verticalPos):
    """Updates the position for the file"""
    if fileName:
        project = GlobalData().project
        if project.isLoaded():
            if project.isProjectFile(fileName):
                key = relpath(fileName, dirname(project.fileName))
                project.updateFilePosition(key, line, pos, firstLine,
                                           horizontalPos, verticalPos)
                return
        Settings().updateFilePosition(fileName, line, pos, firstLine,
                                      horizontalPos, verticalPos)


def getFindFileHistory():
    """Provides the find file history"""
    project = GlobalData().project
    if project.isLoaded():
        return project.findFileHistory
    return Settings().findFileHistory


def setFindFileHistory(values):
    """Updates the find file history"""
    project = GlobalData().project
    if project.isLoaded():
        project.findFileHistory = values
    else:
        Settings().findFileHistory = values


def getFindNameHistory():
    """Provides the find name history"""
    project = GlobalData().project
    if project.isLoaded():
        return project.findNameHistory
    return Settings().findNameHistory


def setFindNameHistory(values):
    """Updates the find name history"""
    project = GlobalData().project
    if project.isLoaded():
        project.findNameHistory = values
    else:
        Settings().findNameHistory = values


def getFindInFilesHistory():
    """Provides the find in files history"""
    project = GlobalData().project
    if project.isLoaded():
        return project.findInFilesHistory
    return Settings().findInFilesHistory


def setFindInFilesHistory(values):
    """Updates the find in files history"""
    project = GlobalData().project
    if project.isLoaded():
        project.findInFilesHistory = values
    else:
        Settings().findInFilesHistory = values


def getFindHistory():
    """Provides the find history"""
    project = GlobalData().project
    if project.isLoaded():
        return project.findHistory
    return Settings().findHistory


def setFindHistory(values):
    """Updates the find history"""
    project = GlobalData().project
    if project.isLoaded():
        project.findHistory = values
    else:
        Settings().findHistory = values


def addRecentFile(path):
    """Adds a recent file"""
    project = GlobalData().project
    if project.isLoaded():
        project.addRecentFile(path)
    else:
        Settings().addRecentFile(path)


def removeRecentFile(path):
    """Removes a recent file"""
    project = GlobalData().project
    if project.isLoaded():
        project.removeRecentFile(path)
    else:
        Settings().removeRecentFile(path)


def getRecentFiles():
    """Provides the recent files list"""
    project = GlobalData().project
    if project.isLoaded():
        return project.recentFiles
    return Settings().recentFiles




##DebuggerEnvironment
##SearchEnvironment
##FileSystemEnvironment



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Process related utility functions"""

import os
import errno
import time
from signal import SIGTERM, SIGINT, SIGHUP, SIGKILL


def isProcessAlive(pid):
    """Returns True is the process is still alive"""
    try:
        # Signal 0 means no sending a signal but check preconditions
        os.kill(pid, 0)
    except OSError as excpt:
        if excpt.errno == errno.ESRCH:
            return False
        if excpt.errno == errno.EPERM:
            return True
        raise
    return True


def killProcess(pid):
    """Tries to kill the given process"""
    for signal in (SIGTERM, SIGINT, SIGHUP):
        if not isProcessAlive(pid):
            return

        try:
            os.kill(pid, signal)
        except OSError as excpt:
            if excpt.errno == errno.ESRCH:
                return  # Already dead
            raise
        time.sleep(0.5)

    # Could not kill gracefully, try harder
    startTime = time.time()
    while True:
        if not isProcessAlive(pid):
            return

        if time.time() - startTime >= 15:
            raise Exception("Cannot kill process (pid: " + str(pid) + ")")

        try:
            os.kill(pid, SIGKILL)
        except OSError as excpt:
            if excpt.errno == errno.ESRCH:
                return  # Already dead
            raise
        time.sleep(0.1)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Codimension main window redirected io support"""

from utils.pixmapcache import getIcon
from utils.runparams import RUN, PROFILE, DEBUG
from .qt import QApplication, QCursor, Qt, QTabBar


class MainWindowRedirectedIOMixin:

    """Main window redirected IO mixin"""

    def __init__(self):
        self.redirectedIOConsole = None
        self.__newRunIndex = -1
        self.__newProfileIndex = -1
        self.__newDebugIndex = -1

    def _initRedirectedIO(self):
        """Connects the signals etc."""
        # The only redirected IO consoles are cloasable
        self._bottomSideBar.sigTabCloseRequested.connect(self.__onCloseRequest)

    def __onCloseRequest(self, index):
        """User wants to close a redirected IO console"""
        self._bottomSideBar.removeTab(index)

    def __getNewRunIndex(self):
        """Provides the new run index"""
        self.__newRunIndex += 1
        return self.__newRunIndex

    def __getNewProfileIndex(self):
        """Provides the new profile index"""
        self.__newProfileIndex += 1
        return self.__newProfileIndex

    def __getNewDebugIndex(self):
        """Provides a new debug console index"""
        self.__newDebugIndex += 1
        return self.__newDebugIndex

    def __getCaptionNameTooltip(self, kind):
        """Provides the tab caption, name and tooltip"""
        if kind == PROFILE:
            index = str(self.__getNewProfileIndex())
            return ('Profiling #' + index, 'profiling#' + index,
                    'Redirected IO profile console #' + index + ' (running)')
        if kind == RUN:
            index = str(self.__getNewRunIndex())
            return ('Run #' + index, 'running#' + index,
                    'Redirected IO run console #' + index + ' (running)')
        index = str(self.__getNewDebugIndex())
        return ('Debug #' + index, 'debugging#' + index,
                'Redirected IO debug console #' + index + ' (running)')

    def addIOConsole(self, widget, consoleType):
        """Installs a new widget at the bottom"""
        if consoleType not in [RUN, PROFILE, DEBUG]:
            raise Exception('Undefined redirected IO console type')

        caption, name, tooltip = self.__getCaptionNameTooltip(consoleType)

        widget.sigKillIOConsoleProcess.connect(self.__onKillIOConsoleProcess)
        widget.sigSettingsUpdated.connect(self.onIOConsoleSettingsUpdated)

        self._bottomSideBar.addTab(
            widget, getIcon('ioconsole.png'), caption, name, None)
        self._bottomSideBar.tabButton(widget, QTabBar.RightSide).hide()
        self._bottomSideBar.setTabToolTip(name, tooltip)
        self._bottomSideBar.show()
        self._bottomSideBar.setCurrentTab(name)
        self._bottomSideBar.raise_()
        widget.setFocus()

    def __onKillIOConsoleProcess(self, procuuid):
        """Kills the process linked to the IO console"""
        self._runManager.kill(procuuid)

    def onIOConsoleSettingsUpdated(self):
        """Initiates updating all the IO consoles settings"""
        index = self._bottomSideBar.count - 1
        while index >= 0:
            widget = self._bottomSideBar.widget(index)
            if hasattr(widget, 'procuuid'):
                if hasattr(widget, 'consoleSettingsUpdated'):
                    widget.consoleSettingsUpdated()
            index -= 1

    def __onClientStdout(self, data):
        """Triggered when the client reports stdout"""
        self._bottomSideBar.show()
        self._bottomSideBar.setCurrentTab('ioredirect')
        self._bottomSideBar.raise_()
        self.redirectedIOConsole.appendStdoutMessage(data)

    def __onClientStderr(self, data):
        """Triggered when the client reports stderr"""
        self._bottomSideBar.show()
        self._bottomSideBar.setCurrentTab('ioredirect')
        self._bottomSideBar.raise_()
        self.redirectedIOConsole.appendStderrMessage(data)

    def __ioconsoleIDEMessage(self, message):
        """Sends an IDE message to the IO console"""
        self._bottomSideBar.show()
        self._bottomSideBar.setCurrentTab('ioredirect')
        self._bottomSideBar.raise_()
        self.redirectedIOConsole.appendIDEMessage(message)

    def __onClientRawInput(self, prompt, echo):
        """Triggered when the client input is requested"""
        self._bottomSideBar.show()
        self._bottomSideBar.setCurrentTab('ioredirect')
        self._bottomSideBar.raise_()
        self.redirectedIOConsole.rawInput(prompt, echo)
        self.redirectedIOConsole.setFocus()

    def __onUserInput(self, userInput):
        """Triggered when the user finished input in the redirected IO tab"""
        self.__debugger.remoteRawInput(userInput)

    def updateIOConsoleTooltip(self, procuuid, msg):
        """Updates the IO console tooltip"""
        index = self.__getIOConsoleIndex(procuuid)
        if index is not None:
            tooltip = self._bottomSideBar.tabToolTip(index)
            tooltip = tooltip.replace("(running)", "(" + msg + ")")
            self._bottomSideBar.setTabToolTip(index, tooltip)

    def __getIOConsoleIndex(self, procuuid):
        """Provides the IO console index by the thread ID"""
        index = self._bottomSideBar.count - 1
        while index >= 0:
            widget = self._bottomSideBar.widget(index)
            if hasattr(widget, "procuuid"):
                if widget.procuuid == procuuid:
                    return index
            index -= 1
        return None

    def __onCloseIOConsole(self, procuuid):
        """Closes the tab with the corresponding widget"""
        index = self.__getIOConsoleIndex(procuuid)
        if index is not None:
            self._bottomSideBar.removeTab(index)

    def closeAllIOConsoles(self):
        """Closes all IO run/profile tabs and clears the debug IO console"""
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        index = self._bottomSideBar.count - 1
        while index >= 0:
            widget = self._bottomSideBar.widget(index)
            if hasattr(widget, 'procuuid'):
                if hasattr(widget, "stopAndClose"):
                    widget.stopAndClose()
            index -= 1
        QApplication.restoreOverrideCursor()

    def getIOConsoles(self):
        """Provides a list of the current IO consoles"""
        consoles = []
        index = self._bottomSideBar.count - 1
        while index >= 0:
            widget = self._bottomSideBar.widget(index)
            if hasattr(widget, 'procuuid'):
                consoles.append(widget)
            index -= 1
        return consoles

    def onReuseConsole(self, widget, kind):
        """Called when a console is reused"""
        caption, name, tooltip = self.__getCaptionNameTooltip(kind)
        self._bottomSideBar.tabButton(widget, QTabBar.RightSide).hide()
        self._bottomSideBar.updateTabName(widget, name)
        self._bottomSideBar.setTabText(widget, caption)
        self._bottomSideBar.setTabToolTip(widget, tooltip)
        self._bottomSideBar.show()
        self._bottomSideBar.setCurrentTab(widget)
        self._bottomSideBar.raise_()
        widget.setFocus()

    def onConsoleFinished(self, widget):
        """Triggered when a process finished one way or another"""
        self._bottomSideBar.tabButton(widget, QTabBar.RightSide).show()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Various routines used in other places"""

# pylint: disable=C0305

from ui.qt import QPainterPath
from .cml import CMLdoc


def distance(val, begin, end):
    """Provides a distance between the absPos and an item"""
    if val >= begin and val <= end:
        return 0
    return min(abs(val - begin), abs(val - end))


def getCommentBoxPath(settings, baseX, baseY, width, height):
    """Provides the comment box path"""
    return getNoCellCommentBoxPath(baseX + settings.hCellPadding,
                                   baseY + settings.vCellPadding,
                                   width - 2 * settings.hCellPadding,
                                   height - 2 * settings.vCellPadding,
                                   settings.commentCorner)


def getNoCellCommentBoxPath(xPos, yPos, width, height, corner):
    """Provides the path for exactly specified rectangle"""
    path = QPainterPath()
    path.moveTo(xPos, yPos)
    path.lineTo(xPos + width - corner, yPos)
    path.lineTo(xPos + width, yPos + corner)
    path.lineTo(xPos + width, yPos + height)
    path.lineTo(xPos, yPos + height)
    path.lineTo(xPos, yPos)

    # -1 is to avoid sharp corners of the lines
    path.moveTo(xPos + width - corner, yPos + 1)
    path.lineTo(xPos + width - corner, yPos + corner)
    path.lineTo(xPos + width - 1, yPos + corner)
    return path


def getCMLComment(cmlComments, code):
    """CML comment or None"""
    for cmlComment in cmlComments:
        if hasattr(cmlComment, 'CODE'):
            if cmlComment.CODE == code:
                return cmlComment
    return None


def getDocComment(cmlComments):
    """CML doc comment reference or None"""
    return getCMLComment(cmlComments, CMLdoc.CODE)




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2011-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""The tag help viewer implementation"""

from utils.pixmapcache import getIcon
from utils.colorfont import getZoomedMonoFont, getLabelStyle, HEADER_HEIGHT
from .qt import (Qt, QSize, QPlainTextEdit, QMenu, QApplication, QHBoxLayout,
                 QWidget, QAction, QToolBar, QSizePolicy, QLabel, QVBoxLayout,
                 QCursor)


class TagHelpViewer(QWidget):

    """The tag help viewer widget"""

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.__isEmpty = True
        self.__copyAvailable = False
        self.__clearButton = None
        self.__textEdit = None
        self.__header = None
        self.__copyButton = None
        self.__selectAllButton = None
        self.__createLayout(parent)

        # create the context menu
        self.__menu = QMenu(self)
        self.__selectAllMenuItem = self.__menu.addAction(
            getIcon('selectall.png'), 'Select All', self.__textEdit.selectAll)
        self.__copyMenuItem = self.__menu.addAction(
            getIcon('copymenu.png'), 'Copy', self.__textEdit.copy)
        self.__menu.addSeparator()
        self.__clearMenuItem = self.__menu.addAction(
            getIcon('trash.png'), 'Clear', self.__clear)

        self.__textEdit.setContextMenuPolicy(Qt.CustomContextMenu)
        self.__textEdit.customContextMenuRequested.connect(
            self.__handleShowContextMenu)
        self.__textEdit.copyAvailable.connect(self.__onCopyAvailable)

        self.__updateToolbarButtons()

    def __createLayout(self, parent):
        """Helper to create the viewer layout"""
        # __textEdit list area
        self.__textEdit = QPlainTextEdit(parent)
        self.__textEdit.setLineWrapMode(QPlainTextEdit.NoWrap)
        self.__textEdit.setFont(getZoomedMonoFont())
        self.__textEdit.setReadOnly(True)

        # Buttons
        self.__selectAllButton = QAction(
            getIcon('selectall.png'), 'Select all', self)
        self.__selectAllButton.triggered.connect(self.__textEdit.selectAll)
        self.__copyButton = QAction(
            getIcon('copymenu.png'), 'Copy to clipboard', self)
        self.__copyButton.triggered.connect(self.__textEdit.copy)
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.__clearButton = QAction(getIcon('trash.png'), 'Clear all', self)
        self.__clearButton.triggered.connect(self.__clear)

        # Toolbar
        toolbar = QToolBar()
        toolbar.setOrientation(Qt.Vertical)
        toolbar.setMovable(False)
        toolbar.setAllowedAreas(Qt.LeftToolBarArea)
        toolbar.setIconSize(QSize(16, 16))
        toolbar.setFixedWidth(28)
        toolbar.setContentsMargins(0, 0, 0, 0)
        toolbar.addAction(self.__selectAllButton)
        toolbar.addAction(self.__copyButton)
        toolbar.addWidget(spacer)
        toolbar.addAction(self.__clearButton)

        self.__header = QLabel("Signature: none", self)
        self.__header.setObjectName('signature')
        self.__header.setStyleSheet('QFrame#signature {' +
                                    getLabelStyle(self.__header) + '}')
        self.__header.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Fixed)
        self.__header.setFixedHeight(HEADER_HEIGHT)
        verticalLayout = QVBoxLayout()
        verticalLayout.setContentsMargins(2, 2, 2, 2)
        verticalLayout.setSpacing(2)
        verticalLayout.addWidget(self.__header)
        verticalLayout.addWidget(self.__textEdit)

        # layout
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(toolbar)
        layout.addLayout(verticalLayout)

        self.setLayout(layout)

    def __handleShowContextMenu(self, _):
        """Show the context menu"""
        self.__selectAllMenuItem.setEnabled(not self.__isEmpty)
        self.__copyMenuItem.setEnabled(self.__copyAvailable)
        self.__clearMenuItem.setEnabled(not self.__isEmpty)
        self.__menu.popup(QCursor.pos())

    @staticmethod
    def __calltipDisplayable(calltip):
        """True if calltip is displayable"""
        if calltip is None:
            return False
        if calltip.strip() == "":
            return False
        return True

    @staticmethod
    def __docstringDisplayable(docstring):
        """True if docstring is displayable"""
        if docstring is None:
            return False
        if isinstance(docstring, dict):
            if docstring["docstring"].strip() == "":
                return False
            return True
        if docstring.strip() == "":
            return False
        return True

    def display(self, calltip, docstring):
        """Displays the given help information"""
        calltipDisplayable = self.__calltipDisplayable(calltip)
        docstringDisplayable = self.__docstringDisplayable(docstring)
        self.__isEmpty = True
        if calltipDisplayable or docstringDisplayable:
            self.__isEmpty = False

        if calltipDisplayable:
            if '\n' in calltip:
                calltip = calltip.split('\n')[0]
            self.__header.setText("Signature: " + calltip.strip())
        else:
            self.__header.setText("Signature: n/a")

        self.__textEdit.clear()
        if docstringDisplayable:
            if isinstance(docstring, dict):
                docstring = docstring["docstring"]
            self.__textEdit.insertPlainText(docstring)

        self.__updateToolbarButtons()
        QApplication.processEvents()

    def __updateToolbarButtons(self):
        """Contextually updates toolbar buttons"""
        self.__selectAllButton.setEnabled(not self.__isEmpty)
        self.__copyButton.setEnabled(self.__copyAvailable)
        self.__clearButton.setEnabled(not self.__isEmpty)

    def __clear(self):
        """Triggers when the clear function is selected"""
        self.__isEmpty = True
        self.__copyAvailable = False
        self.__header.setText("Signature: none")
        self.__textEdit.clear()
        self.__updateToolbarButtons()

    def __onCopyAvailable(self, isAvailable):
        """Triggers on the copyAvailable signal"""
        self.__copyAvailable = isAvailable
        self.__updateToolbarButtons()
